@@ -73,12 +73,15 @@ static GCC_INLINE ub4 perl_hash_function(
         if (condition_expr)     \
         {      \
             fcs_compact_alloc_release(stacks_allocator);    \
-            new_state->s.stacks[i] = cached_stack;       \
+            *(current_stack) = cached_stack;       \
         }
 
-static void GCC_INLINE fc_solve_cache_stacks(
+/* TODO : Maybe define an accesor for new_state_key->stacks (also see the
+ * replaced_with_cached macro above.
+ * */
+static GCC_INLINE void fc_solve_cache_stacks(
         fc_solve_hard_thread_t * hard_thread,
-        fcs_state_keyval_pair_t * new_state
+        fcs_kv_state_t * new_state
         )
 {
     int i;
@@ -98,26 +101,29 @@ static void GCC_INLINE fc_solve_cache_stacks(
     fcs_cards_column_t column;
     register int col_len;
     fcs_compact_allocator_t * stacks_allocator;
+    register fcs_state_t * new_state_key = new_state->key;
+    register fcs_state_extra_info_t * new_state_info = new_state->val;
+    register fcs_cards_column_t * current_stack;
 
     stacks_allocator = &(hard_thread->allocator);
 
-    for(i=0 ; i < LOCAL_STACKS_NUM ; i++)
+    for (i=0, current_stack = new_state_key->stacks ; i < LOCAL_STACKS_NUM ; i++, current_stack++)
     {
         /*
          * If the stack is not a copy - it is already cached so skip
          * to the next stack
          * */
-        if (! (new_state->info.stacks_copy_on_write_flags & (1 << i)))
+        if (! (new_state_info->stacks_copy_on_write_flags & (1 << i)))
         {
             continue;
         }
 
-        column = fcs_state_get_col(new_state->s, i);
+        column = fcs_state_get_col(*(new_state_key), i);
         col_len = (fcs_col_len(column)+1);
 
         new_ptr = (char*)fcs_compact_alloc_ptr(stacks_allocator, col_len);
         memcpy(new_ptr, column, col_len);
-        new_state->s.stacks[i] = new_ptr;
+        *(current_stack) = new_ptr;
 
 #if FCS_STACK_STORAGE == FCS_STACK_STORAGE_INTERNAL_HASH
 #ifdef FCS_ENABLE_SECONDARY_HASH_VALUE
@@ -125,7 +131,7 @@ static void GCC_INLINE fc_solve_cache_stacks(
         /* This hash function was ripped from the Perl source code.
          * (It is not derived work however). */
         {
-            const char * s_ptr = (char*)(new_state_key->stacks[i]);
+            const char * s_ptr = (char*)(*(current_stack));
             const char * s_end = s_ptr+fcs_col_len(s_ptr)+1;
             hash_value_int = 0;
             while (s_ptr < s_end)
@@ -146,14 +152,14 @@ static void GCC_INLINE fc_solve_cache_stacks(
 #endif
 
         {
-            column = fcs_state_get_col(new_state->s, i);
+            column = fcs_state_get_col(*(new_state_key), i);
 
             replace_with_cached(fc_solve_hash_insert(
                 &(instance->stacks_hash),
                 column,
                 &cached_stack,
                 perl_hash_function(
-                    (ub1 *)new_state->s.stacks[i],
+                    (ub1 *)*(current_stack),
                     col_len
                     )
 #ifdef FCS_ENABLE_SECONDARY_HASH_VALUE
@@ -180,22 +186,22 @@ static void GCC_INLINE fc_solve_cache_stacks(
         cached_stack =
             fcs_libavl2_stacks_tree_insert(
                 instance->stacks_tree,
-                new_state_key->stacks[i]
+                new_state->s.stacks[i]
             );
 
         replace_with_cached(cached_stack != NULL);
 
 #elif (FCS_STACK_STORAGE == FCS_STACK_STORAGE_LIBREDBLACK_TREE)
         cached_stack = (void *)rbsearch(
-            new_state_key->stacks[i],
+            *(current_stack),
             instance->stacks_tree
             );
 
-        replace_with_cached(cached_stack != new_state_key->stacks[i]);
+        replace_with_cached(cached_stack != *(current_stack));
 #elif (FCS_STACK_STORAGE == FCS_STACK_STORAGE_GLIB_TREE)
         cached_stack = g_tree_lookup(
              instance->stacks_tree,
-             (gpointer)new_state_key->stacks[i]
+             (gpointer)*(current_stack)
              );
 
         /* replace_with_cached contains an if statement */
@@ -204,22 +210,22 @@ static void GCC_INLINE fc_solve_cache_stacks(
         {
             g_tree_insert(
                 instance->stacks_tree,
-                (gpointer)new_state_key->stacks[i],
-                (gpointer)new_state_key->stacks[i]
+                (gpointer)*(current_stack),
+                (gpointer)*(current_stack)
                 );
         }
 #elif (FCS_STACK_STORAGE == FCS_STACK_STORAGE_GLIB_HASH)
         cached_stack = g_hash_table_lookup(
             instance->stacks_hash,
-            (gconstpointer)new_state_key->stacks[i]
+            (gconstpointer)*(current_stack)
             );
         replace_with_cached(cached_stack != NULL)
         else
         {
             g_hash_table_insert(
                 instance->stacks_hash,
-                (gpointer)new_state_key->stacks[i],
-                (gpointer)new_state_key->stacks[i]
+                (gpointer)*(current_stack),
+                (gpointer)*(current_stack)
                 );
         }
 #elif (FCS_STACK_STORAGE == FCS_STACK_STORAGE_JUDY)
@@ -255,43 +261,6 @@ static void GCC_INLINE fc_solve_cache_stacks(
 
 #endif
 
-
-#ifdef FCS_WITH_TALONS
-static GCC_INLINE void fc_solve_cache_talon(
-    fc_solve_instance_t * instance,
-    fcs_state_t * new_state_key,
-    fcs_state_extra_info_t * new_state_val
-    )
-{
-    void * cached_talon;
-    int hash_value_int;
-
-    new_state_key->talon = realloc(new_state_key->talon, fcs_klondike_talon_len(new_state_key)+1);
-    hash_value_int = *(SFO_hash_value_t*)instance->hash_value;
-    if (hash_value_int < 0)
-    {
-        /*
-         * This is a bit mask that nullifies the sign bit of the
-         * number so it will always be positive
-         * */
-        hash_value_int &= (~(1<<((sizeof(hash_value_int)<<3)-1)));
-    }
-
-    cached_talon = (void *)fc_solve_hash_insert(
-        instance->talons_hash,
-        new_state_key->talon,
-        hash_value_int,
-        1
-        );
-
-    if (cached_talon != NULL)
-    {
-        free(new_state_key->talon);
-        new_state_key->talon = cached_talon;
-    }
-}
-#endif
-
 #if (FCS_STATE_STORAGE == FCS_STATE_STORAGE_GLIB_HASH)
 guint fc_solve_hash_function(gconstpointer key)
 {
@@ -333,20 +302,19 @@ guint fc_solve_hash_function(gconstpointer key)
 static GCC_INLINE void on_state_new(
     fc_solve_instance_t * instance,
     fc_solve_hard_thread_t * hard_thread,
-    fcs_collectible_state_t * new_state
+    fcs_state_extra_info_t * new_state_info
 )
 {
     register fcs_collectible_state_t * parent_state;
-
     /* The new state was not found in the cache, and it was already inserted */
-    if (likely(parent_state = FCS_S_PARENT(new_state)))
+    if (likely(parent_state = new_state_info->parent))
     {
         (FCS_S_NUM_ACTIVE_CHILDREN(parent_state))++;
         /* If parent_val is defined, so is moves_to_parent */
-        FCS_S_MOVES_TO_PARENT(new_state) =
+        new_state_info->moves_to_parent =
             fc_solve_move_stack_compact_allocate(
                 hard_thread,
-                FCS_S_MOVES_TO_PARENT(new_state)
+                new_state_info->moves_to_parent
             );
     }
 
@@ -356,22 +324,22 @@ static GCC_INLINE void on_state_new(
     return;
 }
 
-#ifdef FCS_RCS_STATES
 fcs_bool_t fc_solve_check_and_add_state(
     fc_solve_hard_thread_t * hard_thread,
-    fcs_state_t * new_state_key,
-    fcs_collectible_state_t * new_state,
-    fcs_state_t * * existing_state_key,
-    fcs_collectible_state_t * * existing_state
+    fcs_kv_state_t * new_state,
+    fcs_kv_state_t * existing_state_raw
     )
-#else
-fcs_bool_t fc_solve_check_and_add_state(
-    fc_solve_hard_thread_t * hard_thread,
-    fcs_collectible_state_t * new_state,
-    fcs_collectible_state_t * * existing_state
-    )
-#endif
 {
+/*
+ * TODO : these accessor macros are probably out-of-date and won't work with
+ * some of the less commonly tested data storage backends. They should be
+ * tested and updated.
+ * */
+#define existing_state_val (existing_state_raw->val)
+#define new_state_key      (new_state->key)
+
+#define ON_STATE_NEW() on_state_new(instance, hard_thread, new_state->val);
+
 #if (FCS_STATE_STORAGE == FCS_STATE_STORAGE_INTERNAL_HASH)
 #ifdef FCS_ENABLE_SECONDARY_HASH_VALUE
     SFO_hash_value_t hash_value_int;
@@ -394,15 +362,15 @@ fcs_bool_t fc_solve_check_and_add_state(
 
     fc_solve_cache_stacks(hard_thread, new_state);
 
-    fc_solve_canonize_state(
-#ifdef FCS_RCS_STATES
-            new_state_key,
-#endif
+    {
+        fc_solve_canonize_state(
             new_state,
             INSTANCE_FREECELLS_NUM,
             INSTANCE_STACKS_NUM
             );
 
+    }
+
 /*
     The objective of this part of the code is:
     1. To check if new_state_key / new_state_val is already in the prev_states
@@ -436,31 +404,28 @@ fcs_bool_t fc_solve_check_and_add_state(
         void * existing_void;
         if (fc_solve_hash_insert(
         &(instance->hash),
-        new_state,
 #ifdef FCS_RCS_STATES
-        new_state_key,
+        new_state->val,
+        new_state->key,
+#else
+        FCS_STATE_kv_to_collectible(new_state),
 #endif
         &existing_void,
         perl_hash_function(
-#ifdef FCS_RCS_STATES
             (ub1 *)(new_state_key),
-            sizeof(*new_state_key)
-#else
-            (ub1 *)&(new_state->s),
-            sizeof(new_state->s)
-#endif
+            sizeof(*(new_state_key))
             )
 #ifdef FCS_ENABLE_SECONDARY_HASH_VALUE
         , hash_value_int
 #endif
         ))
         {
-            *existing_state = existing_void;
+            FCS_STATE_collectible_to_kv(existing_state_raw, existing_void);
             return FALSE;
         }
         else
         {
-            on_state_new(instance, hard_thread, new_state);
+            ON_STATE_NEW();
             return TRUE;
         }
     }
@@ -476,12 +441,12 @@ fcs_bool_t fc_solve_check_and_add_state(
             )
         )
         {
-            *existing_state = existing_void;
+            existing_state_val = existing_void;
             return FALSE;
         }
         else
         {
-            on_state_new(instance, hard_thread, new_state);
+            ON_STATE_NEW();
             return TRUE;
         }
     }
@@ -506,7 +471,7 @@ fcs_bool_t fc_solve_check_and_add_state(
         if (found)
         {
             is_state_new = FALSE;
-            *existing_state_val = pos_ptr->val;
+            existing_state_val = pos_ptr->val;
         }
         else
         {
@@ -556,53 +521,69 @@ fcs_bool_t fc_solve_check_and_add_state(
     }
     else
     {
-        *existing_state_val = pos_ptr->val;
+        existing_state_val = pos_ptr->val;
         is_state_new = FALSE;
     }
 
 #elif (FCS_STATE_STORAGE == FCS_STATE_STORAGE_LIBREDBLACK_TREE)
-    *existing_state_val = (fcs_state_extra_info_t *)rbsearch(new_state_val,
+    existing_state_val = (fcs_state_extra_info_t *)rbsearch(new_state_val,
             instance->tree
             );
     is_state_new = ((*existing_state_val) == new_state_val);
 
 #elif (FCS_STATE_STORAGE == FCS_STATE_STORAGE_KAZ_TREE)
 #ifdef FCS_RCS_STATES
-    instance->tree_new_state_key = new_state_key;
-    instance->tree_new_state = new_state;
+    instance->tree_new_state_key = new_state->key;
+    instance->tree_new_state = new_state->val;
 #endif
 
-    if ((*existing_state = (fcs_collectible_state_t *)
-        fc_solve_kaz_tree_alloc_insert(instance->tree, new_state))
+    {
+        void * existing_void;
+    if ((existing_void =
+        fc_solve_kaz_tree_alloc_insert(instance->tree,
+#ifdef FCS_RCS_STATES
+                                       new_state_val
+#else
+                                       FCS_STATE_kv_to_collectible(new_state)
+#endif
+                                       ))
             == NULL
        )
     {
-        on_state_new(instance, hard_thread, new_state);
+        ON_STATE_NEW();
         return TRUE;
     }
     else
     {
+        FCS_STATE_collectible_to_kv(existing_state_raw, existing_void);
         return FALSE;
     }
+    }
 #elif (FCS_STATE_STORAGE == FCS_STATE_STORAGE_LIBAVL2_TREE)
-    instance->tree_new_state_key = new_state_key;
-    instance->tree_new_state = new_state;
+#ifdef FCS_RCS_STATES
+    instance->tree_new_state_key = new_state->key;
+    instance->tree_new_state = new_state->val;
+#endif
 
-    if ((*existing_state = (fcs_collectible_state_t *)
-        fcs_libavl2_states_tree_insert(instance->tree, new_state))
+    {
+        void * existing_void;
+    if ((existing_void =
+        fcs_libavl2_states_tree_insert(instance->tree, FCS_STATE_kv_to_collectible(new_state)))
             == NULL
        )
     {
-        on_state_new(instance, hard_thread, new_state);
+        ON_STATE_NEW();
         return TRUE;
     }
     else
     {
+        FCS_STATE_collectible_to_kv(existing_state_raw, existing_void);
         return FALSE;
     }
+    }
 #elif (FCS_STATE_STORAGE == FCS_STATE_STORAGE_GLIB_TREE)
-    *existing_state_val = g_tree_lookup(instance->tree, (gpointer)new_state_key);
-    if ((is_state_new = (*existing_state_val == NULL)))
+    existing_state_val = g_tree_lookup(instance->tree, (gpointer)new_state_key);
+    if ((is_state_new = (existing_state_val == NULL)))
     {
         /* The new state was not found. Let's insert it.
          * The value must be the same as the key, so g_tree_lookup()
@@ -615,9 +596,9 @@ fcs_bool_t fc_solve_check_and_add_state(
     }
 
 #elif (FCS_STATE_STORAGE == FCS_STATE_STORAGE_GLIB_HASH)
-    *existing_state_val = g_hash_table_lookup(instance->hash,
+    existing_state_val = g_hash_table_lookup(instance->hash,
             (gpointer)new_state_key);
-    if ((is_state_new = (*existing_state_val == NULL)))
+    if ((is_state_new = (existing_state_val == NULL)))
     {
         /* The new state was not found. Let's insert it.
          * The value must be the same as the key, so g_tree_lookup()
@@ -667,12 +648,12 @@ fcs_bool_t fc_solve_check_and_add_state(
         if ((is_state_new = (*PValue == 0)))
         {
             /* A new state. */
-            *PValue = (PWord_t)(*existing_state_val = new_state_val);
+            *PValue = (PWord_t)(existing_state_val = new_state_val);
         }
         else
         {
             /* Already exists. */
-            *existing_state_val = (fcs_state_extra_info_t *)(*PValue);
+            existing_state_val = (fcs_state_extra_info_t *)(*PValue);
         }
     }
 #else
