@@ -470,14 +470,7 @@ fc_solve_instance_t * fc_solve_alloc_instance(void)
     instance->opt_tests_order.num = 0;
     instance->opt_tests_order.tests = NULL;
 
-
-
-#ifdef FCS_WITH_TALONS
-    instance->talon_type = FCS_TALON_NONE;
-#endif
-
     instance->num_hard_threads = 0;
-
     instance->hard_threads = NULL;
 
 #ifndef FCS_FREECELL_ONLY
@@ -902,6 +895,14 @@ static int fc_solve_rcs_states_compare(const void * void_a, const void * void_b,
 
 #endif
 
+#define STATE_STORAGE_TREE_COMPARE() fc_solve_rcs_states_compare
+#define STATE_STORAGE_TREE_CONTEXT() instance
+
+#else
+
+#define STATE_STORAGE_TREE_COMPARE() fc_solve_state_compare_with_context
+#define STATE_STORAGE_TREE_CONTEXT() NULL
+
 #endif
 
 /*
@@ -924,20 +925,18 @@ void fc_solve_start_instance_process_with_board(
             sizeof(*state_copy_ptr)
         );
 
-#ifdef FCS_RCS_STATES
-    fcs_duplicate_state(
-        &(state_copy_ptr->s),
-        &(state_copy_ptr->info),
-        &(init_state->s),
-        &(init_state->info)
+    {
+        fcs_kv_state_t pass_copy, pass_init;
 
-        );
-#else
-    fcs_duplicate_state(state_copy_ptr,
-            init_state
-            );
-#endif
+        pass_copy.key = &(state_copy_ptr->s);
+        pass_copy.val = &(state_copy_ptr->info);
+        pass_init.key = &(init_state->s);
+        pass_init.val = &(init_state->info);
 
+        fcs_duplicate_kv_state( &(pass_copy), &(pass_init) );
+    }
+
+#ifdef INDIRECT_STACK_STATES
     {
         int i;
         char * buffer;
@@ -946,9 +945,10 @@ void fc_solve_start_instance_process_with_board(
 
         for ( i=0 ; i < INSTANCE_STACKS_NUM ; i++ )
         {
-            fcs_copy_stack((*state_copy_ptr), i, buffer);
+            fcs_copy_stack(state_copy_ptr->s, state_copy_ptr->info, i, buffer);
         }
     }
+#endif
 
     /* Initialize the state to be a base state for the game tree */
 #ifndef FCS_WITHOUT_DEPTH_FIELD
@@ -969,25 +969,16 @@ void fc_solve_start_instance_process_with_board(
             );
 #elif (FCS_STATE_STORAGE == FCS_STATE_STORAGE_LIBAVL2_TREE)
 
-#ifdef FCS_RCS_STATES
     instance->tree = fcs_libavl2_states_tree_create(
-            fc_solve_rcs_states_compare,
-            instance,
+            STATE_STORAGE_TREE_COMPARE(),
+            STATE_STORAGE_TREE_CONTEXT(),
             NULL
             );
-#else
-    instance->tree = fcs_libavl2_states_tree_create(fc_solve_state_compare_with_context, NULL, NULL);
-#endif
-
 #elif (FCS_STATE_STORAGE == FCS_STATE_STORAGE_KAZ_TREE)
 
-#ifdef FCS_RCS_STATES
     instance->tree = fc_solve_kaz_tree_create(
-            fc_solve_rcs_states_compare, instance
+            STATE_STORAGE_TREE_COMPARE(), STATE_STORAGE_TREE_CONTEXT()
             );
-#else
-    instance->tree = fc_solve_kaz_tree_create(fc_solve_state_compare_with_context, NULL);
-#endif
 
 #elif (FCS_STATE_STORAGE == FCS_STATE_STORAGE_GLIB_TREE)
     instance->tree = g_tree_new(fc_solve_state_compare);
@@ -1076,19 +1067,6 @@ void fc_solve_start_instance_process_with_board(
 
     /***********************************************/
 
-#ifdef FCS_WITH_TALONS
-    /* Initialize the Talon's Cache */
-    if (instance->talon_type == FCS_TALON_KLONDIKE)
-    {
-        instance->talons_hash = fc_solve_hash_init(
-            fcs_talon_compare_with_context,
-            NULL
-            );
-
-        fc_solve_cache_talon(instance, instance->state_copy_ptr_val);
-    }
-#endif
-
 #if (FCS_STATE_STORAGE == FCS_STATE_STORAGE_DB_FILE)
     /* Not working - ignore */
     db_open(
@@ -1103,24 +1081,16 @@ void fc_solve_start_instance_process_with_board(
 #endif
 
     {
-#ifdef FCS_RCS_STATES
-        fcs_collectible_state_t * no_use_val;
-        fcs_state_t * no_use_key;
-#else
-        fcs_state_keyval_pair_t * no_use;
-#endif
+        fcs_kv_state_t no_use;
+        fcs_kv_state_t pass;
+
+        pass.key = &(state_copy_ptr->s);
+        pass.val = &(state_copy_ptr->info);
 
         fc_solve_check_and_add_state(
             instance->hard_threads,
-#ifdef FCS_RCS_STATES
-            &(state_copy_ptr->s),
-            &(state_copy_ptr->info),
-            &no_use_key,
-            &no_use_val
-#else
-            state_copy_ptr,
+            &pass,
             &no_use
-#endif
             );
 
     }
@@ -1221,6 +1191,8 @@ void fc_solve_finish_instance(
 #endif
 #endif
 
+    instance->num_states_in_collection = 0;
+
 #if (FCS_STATE_STORAGE == FCS_STATE_STORAGE_DB_FILE)
     instance->db->close(instance->db,0);
 #endif
