@@ -42,6 +42,7 @@ extern "C" {
 #include "inline.h"
 #include "unused.h"
 
+#include "indirect_buffer.h"
 #include "rand.h"
 
 #if (FCS_STATE_STORAGE == FCS_STATE_STORAGE_LIBREDBLACK_TREE) || (defined(INDIRECT_STACK_STATES) && (FCS_STACK_STORAGE == FCS_STACK_STORAGE_LIBREDBLACK_TREE))
@@ -149,10 +150,7 @@ struct fc_solve_soft_thread_struct;
 
 typedef void (*fc_solve_solve_for_state_test_t)(
         struct fc_solve_soft_thread_struct *,
-#ifdef FCS_RCS_STATES
-        fcs_state_t *,
-#endif
-        fcs_collectible_state_t *,
+        fcs_kv_state_t *,
         fcs_derived_states_list_t *
         );
 
@@ -383,10 +381,7 @@ typedef void (*fcs_instance_debug_iter_output_func_t)(
         int iter_num,
         int depth,
         void * instance,
-#ifdef FCS_RCS_STATES
-        fcs_state_t * state_key,
-#endif
-        fcs_collectible_state_t * state_val,
+        fcs_kv_state_t * state,
         int parent_iter_num
         );
 
@@ -486,26 +481,6 @@ struct fc_solve_instance_struct
      * */
 #if (FCS_STATE_STORAGE == FCS_STATE_STORAGE_DB_FILE)
     DB * db;
-#endif
-#ifdef FCS_WITH_TALONS
-    /*
-     * The talon for Gypsy-like games. Since only the position changes from
-     * state to state.
-     * We can keep it here.
-     *
-     * */
-    fcs_card_t * gypsy_talon;
-
-    /*
-     * The length of the Gypsy talon
-     * */
-    int gypsy_talon_len;
-
-    int talon_type;
-
-    /* The Klondike Talons' Cache */
-    fc_solve_hash_t * talons_hash;
-
 #endif
 
     fcs_runtime_flags_t runtime_flags;
@@ -689,9 +664,7 @@ struct fc_solve_hard_thread_struct
      * This is a buffer used to temporarily store the stacks of the duplicated
      * state.
      * */
-    char indirect_stacks_buffer[MAX_NUM_STACKS << 7];
-#else
-    char indirect_stacks_buffer[1];
+    dll_ind_buf_t indirect_stacks_buffer;
 #endif
 
     int prelude_num_items;
@@ -700,7 +673,7 @@ struct fc_solve_hard_thread_struct
 
     fcs_bool_t allocated_from_list;
     int num_soft_threads;
- 
+
     /*
      * A counter that determines how many of the soft threads that belong
      * to this hard thread have already finished. If it becomes num_soft_threads
@@ -1012,11 +985,9 @@ static GCC_INLINE void fc_solve_soft_thread_init_soft_dfs(
 
     fc_solve_increase_dfs_max_depth(soft_thread);
 
-#ifdef FCS_RCS_STATES
-    soft_thread->method_specific.soft_dfs.soft_dfs_info[0].state = &(ptr_orig_state->info);
-#else
-    soft_thread->method_specific.soft_dfs.soft_dfs_info[0].state = ptr_orig_state;
-#endif
+    soft_thread->method_specific.soft_dfs.soft_dfs_info[0].state
+        = FCS_STATE_keyval_pair_to_collectible(ptr_orig_state);
+
     fc_solve_rand_init(
             &(soft_thread->method_specific.soft_dfs.rand_gen),
             soft_thread->method_specific.soft_dfs.rand_seed
@@ -1367,7 +1338,7 @@ static GCC_INLINE int fc_solve_optimize_solution(
     )
 {
     fc_solve_soft_thread_t * soft_thread;
-    fc_solve_hard_thread_t * optimization_thread;
+    fc_solve_hard_thread_t * old_hard_thread, * optimization_thread;
 
     STRUCT_TURN_ON_FLAG(instance, FCS_RUNTIME_TO_REPARENT_STATES_REAL);
 
@@ -1379,14 +1350,21 @@ static GCC_INLINE int fc_solve_optimize_solution(
 
         fc_solve_instance__init_hard_thread(instance, optimization_thread);
 
+        old_hard_thread = instance->current_hard_thread;
+
+        soft_thread = optimization_thread->soft_threads;
+
+        /* Copy enable_pruning from the thread that reached the solution,
+         * because otherwise -opt in conjunction with -sp r:tf will fail.
+         * */
+        soft_thread->enable_pruning = old_hard_thread->soft_threads[old_hard_thread->st_idx].enable_pruning;
     }
     else
     {
         optimization_thread = instance->optimization_thread;
+        soft_thread = optimization_thread->soft_threads;
     }
 
-    soft_thread = optimization_thread->soft_threads;
-
     if (STRUCT_QUERY_FLAG(instance, FCS_RUNTIME_OPT_TESTS_ORDER_WAS_SET))
     {
         if (soft_thread->by_depth_tests_order.by_depth_tests != NULL)
@@ -1448,7 +1426,7 @@ static GCC_INLINE int fc_solve_resume_instance(
     }
     else
     {
-        fc_solve_hard_thread_t * end_of_hard_threads = 
+        fc_solve_hard_thread_t * end_of_hard_threads =
             instance->hard_threads + instance->num_hard_threads
             ;
 
@@ -1564,11 +1542,11 @@ static GCC_INLINE fc_solve_soft_thread_t *
 }
 
 extern fc_solve_soft_thread_t * fc_solve_new_soft_thread(
-    fc_solve_hard_thread_t * hard_thread
+    fc_solve_hard_thread_t * const hard_thread
     );
 
 static GCC_INLINE fc_solve_soft_thread_t * fc_solve_new_hard_thread(
-    fc_solve_instance_t * instance
+    fc_solve_instance_t * const instance
     )
 {
     fc_solve_hard_thread_t * ret;
@@ -1612,7 +1590,7 @@ static GCC_INLINE fc_solve_soft_thread_t * fc_solve_new_hard_thread(
 /* This is the commmon code from fc_solve_instance__init_hard_thread() and
  * recycle_hard_thread() */
 static GCC_INLINE void fc_solve_reset_hard_thread(
-    fc_solve_hard_thread_t * hard_thread
+    fc_solve_hard_thread_t * const hard_thread
     )
 {
     hard_thread->num_times = 0;
@@ -1621,7 +1599,7 @@ static GCC_INLINE void fc_solve_reset_hard_thread(
 }
 
 static GCC_INLINE void fc_solve_reset_soft_thread(
-    fc_solve_soft_thread_t * soft_thread
+    fc_solve_soft_thread_t * const soft_thread
     )
 {
     STRUCT_CLEAR_FLAG(soft_thread, FCS_SOFT_THREAD_IS_FINISHED);
@@ -1629,8 +1607,8 @@ static GCC_INLINE void fc_solve_reset_soft_thread(
 }
 
 static GCC_INLINE void fc_solve_release_tests_list(
-    fc_solve_soft_thread_t * soft_thread,
-    fcs_bool_t is_scan_befs_or_bfs
+    fc_solve_soft_thread_t * const soft_thread,
+    const fcs_bool_t is_scan_befs_or_bfs
 )
 {
     if (is_scan_befs_or_bfs)
@@ -1668,7 +1646,7 @@ static GCC_INLINE void fc_solve_release_tests_list(
 }
 
 static GCC_INLINE void fc_solve_instance__recycle_hard_thread(
-    fc_solve_hard_thread_t * hard_thread
+    fc_solve_hard_thread_t * const hard_thread
     )
 {
     ST_LOOP_DECLARE_VARS();
@@ -1692,7 +1670,7 @@ static GCC_INLINE void fc_solve_instance__recycle_hard_thread(
 }
 
 static GCC_INLINE void fc_solve_recycle_instance(
-    fc_solve_instance_t * instance
+    fc_solve_instance_t * const instance
         )
 {
     int ht_idx;
@@ -1730,22 +1708,12 @@ extern int fc_solve_compare_lru_cache_keys(
 
 #endif
 
-#ifdef FCS_RCS_STATES
 #define DECLARE_MOVE_FUNCTION(name) \
 extern void name( \
-        fc_solve_soft_thread_t * soft_thread, \
-        fcs_state_t * key_ptr_state_key, \
-        fcs_collectible_state_t * val_ptr_state_val,  \
-        fcs_derived_states_list_t * derived_states_list \
-)
-#else
-#define DECLARE_MOVE_FUNCTION(name) \
-extern void name( \
-        fc_solve_soft_thread_t * soft_thread, \
-        fcs_state_keyval_pair_t * ptr_state, \
-        fcs_derived_states_list_t * derived_states_list \
+        fc_solve_soft_thread_t * const soft_thread, \
+        fcs_kv_state_t * const raw_ptr_state_raw, \
+        fcs_derived_states_list_t * const derived_states_list \
         )
-#endif
 
 #ifdef __cplusplus
 }
