@@ -34,7 +34,9 @@
  *
  * Also see "man assert".
  * */
+#ifndef NDEBUG
 #define NDEBUG
+#endif
 
 #include <stdlib.h>
 #include <stddef.h>
@@ -62,7 +64,6 @@
 #define right dict_right
 #define parent dict_parent
 #define color dict_color
-#define key dict_key
 #define data dict_data
 
 #define nilnode dict_nilnode
@@ -172,7 +173,7 @@ static int verify_bintree(dict_t *dict)
 {
     dnode_t *first, *next;
 
-    first = dict_first(dict);
+    first = fc_solve_kaz_tree_first(dict);
 
     if (dict->dupes) {
         while (first && (next = dict_next(dict, first))) {
@@ -322,15 +323,15 @@ static void safe_traverse(dict_t *dict, void (*func)(dnode_t *, void *))
         case from_parent:
             if (current->left != nil) {
                 next = current->left;
-            } else 
+            } else
         case from_left:
             if (current->right != nil) {
                 came_from = from_parent;
                 next = current->right;
-            } else 
+            } else
         case from_right:
             {
-                came_from = (current == current->parent->left) 
+                came_from = (current == current->parent->left)
                             ? from_left : from_right;
                 next = current->parent;
                 func(current, dict->context);
@@ -449,9 +450,7 @@ void dict_init_like(dict_t *dict, const dict_t *orig)
 
     assert (dict_similar(dict, orig));
 }
-#endif
 
-#ifdef NO_FC_SOLVE
 /*
  * Initialize with allocator
  */
@@ -477,9 +476,7 @@ extern dict_t *dict_init_alloc(dict_t *dict, dictcount_t maxcount,
 #endif
     return dict;
 }
-#endif
 
-#ifdef NO_FC_SOLVE
 /*
  * Remove all nodes from the dictionary (without freeing them in any way).
  */
@@ -495,10 +492,8 @@ static void dict_clear(dict_t *dict)
     assert (dict->nilnode.color == dnode_black);
 }
 
-#endif
 
 
-#ifdef NO_FC_SOLVE
 /*
  * Verify the integrity of the dictionary structure.  This is provided for
  * debugging purposes, and should be placed in assert statements.   Just because
@@ -530,9 +525,7 @@ int dict_verify(dict_t *dict)
         return 0;
     return 1;
 }
-#endif
 
-#ifdef NO_FC_SOLVE
 /*
  * Determine whether two dictionaries are similar: have the same comparison and
  * allocator functions, and same status as to whether duplicates are allowed.
@@ -559,6 +552,15 @@ int dict_similar(const dict_t *left, const dict_t *right)
 }
 #endif
 
+dict_key_t fc_solve_kaz_tree_lookup_value(dict_t *dict, dict_key_t key)
+{
+    dnode_t *dn;
+
+    dn = fc_solve_kaz_tree_lookup(dict, key);
+
+    return (dn ? dn->dict_key : NULL);
+}
+
 /*
  * Locate a node in the dictionary having the given key.
  * If the node is not found, a null a pointer is returned (rather than
@@ -566,7 +568,7 @@ int dict_similar(const dict_t *left, const dict_t *right)
  * located node is returned.
  */
 
-dnode_t *fc_solve_kaz_tree_lookup(dict_t *dict, const void *key)
+dnode_t *fc_solve_kaz_tree_lookup(dict_t *dict, dict_key_t key)
 {
     dnode_t *root = dict_root(dict);
     dnode_t *nil = dict_nil(dict);
@@ -578,7 +580,12 @@ dnode_t *fc_solve_kaz_tree_lookup(dict_t *dict, const void *key)
     /* simple binary search adapted for trees that contain duplicate keys */
 
     while (root != nil) {
-        result = dict->compare(key, root->key, dict->context);
+#ifdef FCS_KAZ_TREE_USE_RECORD_DICT_KEY
+        result = memcmp(&(key.key), &(root->dict_key.key), sizeof(key.key));
+#else
+        result = dict->compare(key, root->dict_key, dict->context);
+#endif
+
         if (result < 0)
             root = root->left;
         else if (result > 0)
@@ -734,19 +741,19 @@ dnode_t *dict_strict_upper_bound(dict_t *dict, const void *key)
  *
  * If a matching key was found - return it. Else - return NULL to indicate
  * that the new key was added.
- * 
+ *
  * The behavior is undefined if the user attempts to insert into
  * a dictionary that is already full (for which the dict_isfull()
  * function returns true).
  */
 
-const void * fc_solve_kaz_tree_insert(dict_t *dict, dnode_t *node, const void *key)
+dict_ret_key_t fc_solve_kaz_tree_insert(dict_t *dict, dnode_t *node, dict_key_t key)
 {
     dnode_t *where = dict_root(dict), *nil = dict_nil(dict);
     dnode_t *parent = nil, *uncle, *grandpa;
     int result = -1;
 
-    node->key = key;
+    node->dict_key = key;
 
     assert (!dict_isfull(dict));
     assert (!dict_contains(dict, node));
@@ -756,7 +763,11 @@ const void * fc_solve_kaz_tree_insert(dict_t *dict, dnode_t *node, const void *k
 
     while (where != nil) {
         parent = where;
-        result = dict->compare(key, where->key, dict->context);
+#ifdef FCS_KAZ_TREE_USE_RECORD_DICT_KEY
+        result = memcmp(&(key.key), &(where->dict_key.key), sizeof(key.key));
+#else
+        result = dict->compare(key, where->dict_key, dict->context);
+#endif
 
         /* We are remming it out because instead of duplicating the key
          * we return the existing key. -- Shlomi Fish, fc-solve.
@@ -767,7 +778,11 @@ const void * fc_solve_kaz_tree_insert(dict_t *dict, dnode_t *node, const void *k
 #endif
         if (result == 0)
         {
+#ifdef FCS_KAZ_TREE_USE_RECORD_DICT_KEY
+            return &(where->dict_key);
+#else
             return where->dict_key;
+#endif
         }
         else if (result < 0)
         {
@@ -1047,14 +1062,14 @@ static GCC_INLINE dnode_t *dnode_init(dnode_t *dnode)
  */
 
 #ifdef NO_FC_SOLVE
-const void * fc_solve_kaz_tree_alloc_insert(dict_t *dict, const void *key, void *data)
+const dict_ret_key_t fc_solve_kaz_tree_alloc_insert(dict_t *dict, const void *key, void *data)
 #else
-const void * fc_solve_kaz_tree_alloc_insert(dict_t *dict, const void *key)
+const dict_ret_key_t fc_solve_kaz_tree_alloc_insert(dict_t *dict, dict_key_t key)
 #endif
 {
     dnode_t * from_bin;
     dnode_t * node;
-    const void * ret;
+    dict_ret_key_t ret;
     fcs_compact_allocator_t * allocator;
 
 
@@ -1103,13 +1118,7 @@ void fc_solve_kaz_tree_delete_free(dict_t *dict, dnode_t *node)
 #endif
 }
 
-#ifdef NO_FC_SOLVE
-/*
- * Return the node with the lowest (leftmost) key. If the dictionary is empty
- * (that is, dict_isempty(dict) returns 1) a null pointer is returned.
- */
-
-dnode_t *dict_first(dict_t *dict)
+dnode_t *fc_solve_kaz_tree_first(dict_t *dict)
 {
     dnode_t *nil = dict_nil(dict), *root = dict_root(dict), *left;
 
@@ -1120,6 +1129,13 @@ dnode_t *dict_first(dict_t *dict)
     return (root == nil) ? NULL : root;
 }
 
+#ifdef NO_FC_SOLVE
+/*
+ * Return the node with the lowest (leftmost) key. If the dictionary is empty
+ * (that is, dict_isempty(dict) returns 1) a null pointer is returned.
+ */
+
+
 /*
  * Return the node with the highest (rightmost) key. If the dictionary is empty
  * (that is, dict_isempty(dict) returns 1) a null pointer is returned.
@@ -1194,12 +1210,6 @@ dnode_t *dict_prev(dict_t *dict, dnode_t *curr)
 }
 #endif
 
-#ifdef NO_FC_SOLVE
-void dict_allow_dupes(dict_t *dict)
-{
-    dict->dupes = 1;
-}
-#endif
 
 #undef dict_count
 #undef dict_isempty
@@ -1209,6 +1219,11 @@ void dict_allow_dupes(dict_t *dict)
 #undef dnode_getkey
 
 #ifdef NO_FC_SOLVE
+void dict_allow_dupes(dict_t *dict)
+{
+    dict->dupes = 1;
+}
+
 dictcount_t dict_count(dict_t *dict)
 {
     return dict->nodecount;
@@ -1223,16 +1238,12 @@ int dict_isfull(dict_t *dict)
 {
     return dict->nodecount == dict->maxcount;
 }
-#endif
 
-#ifdef NO_FC_SOLVE
 int dict_contains(dict_t *dict, dnode_t *node)
 {
     return verify_dict_has_node(dict_nil(dict), dict_root(dict), node);
 }
-#endif
 
-#ifdef NO_FC_SOLVE
 static dnode_t *dnode_alloc(void *context)
 {
     return (dnode_t *) malloc(sizeof *dnode_alloc(NULL));
@@ -1242,9 +1253,7 @@ static void dnode_free(dnode_t *node, void *context)
 {
     free(node);
 }
-#endif
 
-#ifdef NO_FC_SOLVE
 dnode_t *dnode_create(void *data)
 {
     dnode_t *dnode = (dnode_t *) malloc(sizeof *dnode);
@@ -1277,19 +1286,15 @@ void dnode_put(dnode_t *dnode, void *data)
 {
     dnode->data = data;
 }
-#endif
 
-#ifdef NO_FC_SOLVE
 int dnode_is_in_a_dict(dnode_t *dnode)
 {
     return (dnode->parent && dnode->left && dnode->right);
 }
-#endif
 
-#ifdef NO_FC_SOLVE
 void dict_process(dict_t *dict, void *context, dnode_process_t function)
 {
-    dnode_t *node = dict_first(dict), *next;
+    dnode_t *node = fc_solve_kaz_tree_first(dict), *next;
 
     while (node != NULL) {
         /* check for callback function deleting */
@@ -1426,7 +1431,7 @@ void dict_load_end(dict_load_t *load)
 void dict_merge(dict_t *dest, dict_t *source)
 {
     dict_load_t load;
-    dnode_t *leftnode = dict_first(dest), *rightnode = dict_first(source);
+    dnode_t *leftnode = fc_solve_kaz_tree_first(dest), *rightnode = fc_solve_kaz_tree_first(source);
 
     assert (dict_similar(dest, source));
 
@@ -1742,7 +1747,7 @@ int main(void)
                 printf("%lu\n", (unsigned long) dict_count(d));
                 break;
             case 't':
-                for (dn = dict_first(d); dn; dn = dict_next(d, dn)) {
+                for (dn = fc_solve_kaz_tree_first(d); dn; dn = dict_next(d, dn)) {
                     printf("%s\t%s\n", (char *) dnode_getkey(dn),
                             (char *) dnode_get(dn));
                 }
