@@ -38,9 +38,16 @@
 #include "preset.h"
 #include "fcs_user.h"
 #include "move_funcs_order.h"
+#include "fcs_user_internal.h"
 
 #include "unused.h"
 #include "bool.h"
+#include "indirect_buffer.h"
+
+typedef struct {
+    int num_times;
+    int num_states_in_collection;
+} fcs_stats_t;
 
 /* A flare is an alternative scan algorithm to be tried. All flares in
  * a single instance are being evaluated and then one picks the shortest
@@ -112,22 +119,20 @@ typedef struct
     /*
      * The number of iterations this board started at.
      * */
-    int iterations_board_started_at;
+    fcs_stats_t iterations_board_started_at;
     /*
      * The number of iterations that the current instance started solving from.
      * */
-    int init_num_times;
+    fcs_stats_t init_num_times;
     /*
      * A pointer to the currently active instance out of the sequence
      * */
     fc_solve_instance_t * fc_solve_obj;
     fcs_state_keyval_pair_t state;
     fcs_state_keyval_pair_t running_state;
-#ifdef FCS_WITHOUT_LOCS_FIELDS
     fcs_state_locs_struct_t state_locs;
     fcs_state_locs_struct_t trace_solution_state_locs;
     fcs_state_locs_struct_t initial_state_locs;
-#endif
     int ret_code;
     fcs_bool_t all_instances_were_suspended;
     int state_validity_ret;
@@ -137,9 +142,7 @@ typedef struct
 
     fc_solve_soft_thread_t * soft_thread;
 
-#ifdef INDIRECT_STACK_STATES
-    char indirect_stacks_buffer[MAX_NUM_STACKS << 7];
-#endif
+    DECLARE_IND_BUF_T(indirect_stacks_buffer)
     char * state_string_copy;
 
 #ifndef FCS_FREECELL_ONLY
@@ -155,10 +158,7 @@ static void iter_handler_wrapper(
     int iter_num,
     int depth,
     void * lp_instance GCC_UNUSED,
-#ifdef FCS_RCS_STATES
-    fcs_state_t * ptr_state_key,
-#endif
-    fcs_collectible_state_t * ptr_state_val,
+    fcs_kv_state_t * ptr_state,
     int parent_iter_num
     );
 
@@ -212,7 +212,8 @@ static void user_initialize(
     user->current_iterations_limit = -1;
 
     user->state_string_copy = NULL;
-    user->iterations_board_started_at = 0;
+    user->iterations_board_started_at.num_times = 0;
+    user->iterations_board_started_at.num_states_in_collection = 0;
     user->all_instances_were_suspended = TRUE;
 
     user->error_string = NULL;
@@ -260,7 +261,7 @@ int DLLEXPORT freecell_solver_user_apply_preset(
 
     FLARES_LOOP_START()
         status = fc_solve_apply_preset_by_ptr(
-            instance_item->flares[flare_idx].obj,
+            flare->obj,
             new_preset_ptr
             );
 
@@ -312,14 +313,11 @@ int DLLEXPORT freecell_solver_user_set_depth_tests_order(
     )
 {
     fcs_user_t * user;
-    fcs_instance_item_t * instance_item;
     int depth_idx;
     int ret_code;
 
     user = (fcs_user_t *)api_instance;
 
-    instance_item = &(user->instances_list[user->current_instance_idx]);
-
     *error_string = NULL;
 
     if (min_depth < 0)
@@ -384,6 +382,15 @@ int DLLEXPORT freecell_solver_user_set_depth_tests_order(
             error_string
             );
 
+    {
+        int further_depth_idx;
+        for (further_depth_idx = depth_idx+1; further_depth_idx < user->soft_thread->by_depth_tests_order.num ; further_depth_idx++)
+        {
+            free(user->soft_thread->by_depth_tests_order.by_depth_tests[further_depth_idx].tests_order.tests);
+            user->soft_thread->by_depth_tests_order.by_depth_tests[further_depth_idx].tests_order.tests = NULL;
+        }
+    }
+
     user->soft_thread->by_depth_tests_order.by_depth_tests =
         realloc(
             user->soft_thread->by_depth_tests_order.by_depth_tests,
@@ -766,7 +773,8 @@ static void recycle_instance(
              * not get initialized again, and now the num_times of the instance
              * is equal to 0.
              * */
-            user->init_num_times = 0;
+            user->init_num_times.num_times = 0;
+            user->init_num_times.num_states_in_collection = 0;
 
             flare->ret_code = FCS_STATE_NOT_BEGAN_YET;
         }
@@ -789,7 +797,7 @@ int DLLEXPORT freecell_solver_user_resume_solution(
     void * api_instance
     )
 {
-    int init_num_times;
+    fcs_stats_t init_num_times;
     int run_for_first_iteration = 1;
     int ret;
     fcs_user_t * user;
@@ -876,6 +884,7 @@ int DLLEXPORT freecell_solver_user_resume_solution(
         if (flare->ret_code == FCS_STATE_NOT_BEGAN_YET)
         {
             int status;
+            fcs_kv_state_t state_pass;
 
 #if (!(defined(HARD_CODED_NUM_FREECELLS) && defined(HARD_CODED_NUM_STACKS) && defined(HARD_CODED_NUM_DECKS)))
             fc_solve_instance_t * instance = user->fc_solve_obj;
@@ -886,13 +895,8 @@ int DLLEXPORT freecell_solver_user_resume_solution(
                 &(user->state),
                 INSTANCE_FREECELLS_NUM,
                 INSTANCE_STACKS_NUM,
-                INSTANCE_DECKS_NUM
-#ifdef FCS_WITH_TALONS
-                ,user->fc_solve_obj->talon_type
-#endif
-#ifdef INDIRECT_STACK_STATES
-                ,user->indirect_stacks_buffer
-#endif
+                INSTANCE_DECKS_NUM,
+                user->indirect_stacks_buffer
                 );
 
             if (status != FCS_USER_STATE_TO_C__SUCCESS)
@@ -907,9 +911,6 @@ int DLLEXPORT freecell_solver_user_resume_solution(
                 INSTANCE_FREECELLS_NUM,
                 INSTANCE_STACKS_NUM,
                 INSTANCE_DECKS_NUM,
-#ifdef FCS_WITH_TALONS
-                FCS_TALON_NONE,
-#endif
                 &(user->state_validity_card));
 
             if (user->state_validity_ret != FCS_STATE_VALIDITY__OK)
@@ -918,61 +919,31 @@ int DLLEXPORT freecell_solver_user_resume_solution(
                 return user->ret_code;
             }
 
-#ifdef FCS_WITHOUT_LOCS_FIELDS
-            {
-                int i;
-                for ( i=0 ; i<MAX_NUM_STACKS ; i++)
-                {
-                    user->initial_state_locs.stack_locs[i] = (fcs_locs_t)i;
-                }
-                for ( i=0 ; i<MAX_NUM_FREECELLS ; i++)
-                {
-                    user->initial_state_locs.fc_locs[i] = (fcs_locs_t)i;
-                }
-            }
+            fc_solve_init_locs(&(user->initial_state_locs));
             user->state_locs = user->initial_state_locs;
-#endif
 
-            /* running_state is a normalized state. So I'm duplicating
-             * state to it before state is canonized
+            state_pass.key = &(user->state.s);
+            state_pass.val = &(user->state.info);
+            /* running_state is a normalized state. So We're duplicating
+             * state to it before state state_pass is canonized
              * */
-            fcs_duplicate_state(
-#ifdef FCS_RCS_STATES
-                &(user->running_state.s),
-                &(user->running_state.info),
-                &(user->state.s),
-                &(user->state.info)
-#else
-                &(user->running_state),
-                &(user->state)
-#endif
-                );
+            {
+                fcs_kv_state_t pass;
+                pass.key = &(user->running_state.s);
+                pass.val = &(user->running_state.info);
 
-#ifdef FCS_WITHOUT_LOCS_FIELDS
-#endif
+                fcs_duplicate_kv_state(&pass, &state_pass);
+            }
 
-#ifdef FCS_WITHOUT_LOCS_FIELDS
             fc_solve_canonize_state_with_locs
-#else
-            fc_solve_canonize_state
-#endif
                 (
-#ifdef FCS_RCS_STATES
-                &(user->state.s),
-                &(user->state.info),
-#else
-                &(user->state),
-#endif
-#ifdef FCS_WITHOUT_LOCS_FIELDS
+                 &state_pass,
                 &(user->state_locs),
-#endif
                 INSTANCE_FREECELLS_NUM,
                 INSTANCE_STACKS_NUM
                 );
 
-#ifdef FCS_WITHOUT_LOCS_FIELDS
             user->trace_solution_state_locs = user->state_locs;
-#endif
 
             fc_solve_init_instance(user->fc_solve_obj);
 
@@ -980,7 +951,7 @@ int DLLEXPORT freecell_solver_user_resume_solution(
         }
 
 #define parameterized_fixed_limit(increment) \
-    (user->iterations_board_started_at + increment)
+    (user->iterations_board_started_at.num_times + increment)
 #define parameterized_limit(increment) (((increment) < 0) ? (-1) : parameterized_fixed_limit(increment))
 #define local_limit()  \
         (instance_item->limit)
@@ -1017,11 +988,12 @@ int DLLEXPORT freecell_solver_user_resume_solution(
             {
                 user->fc_solve_obj->max_num_times =
                     user->fc_solve_obj->effective_max_num_times =
-                    (user->fc_solve_obj->num_times + mymin - user->iterations_board_started_at);
+                    (user->fc_solve_obj->num_times + mymin - user->iterations_board_started_at.num_times);
             }
         }
 
-        user->init_num_times = init_num_times = user->fc_solve_obj->num_times;
+        user->init_num_times.num_times = init_num_times.num_times = user->fc_solve_obj->num_times;
+        user->init_num_times.num_states_in_collection = init_num_times.num_states_in_collection = user->fc_solve_obj->num_states_in_collection;
 
         if (solve_start)
         {
@@ -1038,38 +1010,36 @@ int DLLEXPORT freecell_solver_user_resume_solution(
             user->all_instances_were_suspended = FALSE;
         }
 
-        user->iterations_board_started_at += user->fc_solve_obj->num_times - init_num_times;
-        user->init_num_times = user->fc_solve_obj->num_times;
+        user->iterations_board_started_at.num_times += user->fc_solve_obj->num_times - init_num_times.num_times;
+        user->iterations_board_started_at.num_states_in_collection += user->fc_solve_obj->num_states_in_collection - init_num_times.num_states_in_collection;
+        user->init_num_times.num_times = user->fc_solve_obj->num_times;
+        user->init_num_times.num_states_in_collection = user->fc_solve_obj->num_states_in_collection;
 
         if (user->ret_code == FCS_STATE_WAS_SOLVED)
         {
+            fcs_kv_state_t pass;
+
 #if (!(defined(HARD_CODED_NUM_FREECELLS) && defined(HARD_CODED_NUM_STACKS) && defined(HARD_CODED_NUM_DECKS)))
             fc_solve_instance_t * instance =
                 user->fc_solve_obj;
 #endif
+
+            pass.key = &(user->state.s);
+            pass.val = &(user->state.info);
             /*
              * TODO : maybe only normalize the final moves' stack in
              * order to speed things up.
              * */
             fc_solve_move_stack_normalize(
                 &(user->fc_solve_obj->solution_moves),
-#ifdef FCS_RCS_STATES
-                &(user->state.s),
-                &(user->state.info),
-#else
-                &(user->state),
-#endif
-#ifdef FCS_WITHOUT_LOCS_FIELDS
+                &(pass),
                 &(user->trace_solution_state_locs),
-#endif
                 INSTANCE_FREECELLS_NUM,
                 INSTANCE_STACKS_NUM,
                 INSTANCE_DECKS_NUM
                 );
 
-#ifdef FCS_WITHOUT_LOCS_FIELDS
             user->trace_solution_state_locs = user->state_locs;
-#endif
 
             if (instance_item->minimal_solution_flare_idx < 0)
             {
@@ -1092,14 +1062,13 @@ int DLLEXPORT freecell_solver_user_resume_solution(
              * and return now.
              * */
             if (((user->current_iterations_limit >= 0) &&
-                (user->iterations_board_started_at >=
+                (user->iterations_board_started_at.num_times >=
                     user->current_iterations_limit)) ||
                 (user->fc_solve_obj->num_states_in_collection >=
                     user->fc_solve_obj->effective_max_num_states_in_collection)
             )
             {
-                /* 
-                 * Bug fix:
+                /* Bug fix:
                  * We need to resume from the last flare in case we exceed
                  * the board iterations limit.
                  * */
@@ -1171,16 +1140,13 @@ int DLLEXPORT freecell_solver_user_get_next_move(
 
             if (ret == 0)
             {
+                fcs_kv_state_t pass;
+                pass.key = &(user->running_state.s);
+                pass.val = &(user->running_state.info);
+
                 fc_solve_apply_move(
-#ifdef FCS_RCS_STATES
-                    &(user->running_state.s),
-                    &(user->running_state.info),
-#else
-                    &(user->running_state),
-#endif
-#ifdef FCS_WITHOUT_LOCS_FIELDS
+                    &(pass),
                     NULL,
-#endif
 #ifdef FCS_USE_COMPACT_MOVE_TOKENS
                     internal_move,
 #else
@@ -1218,15 +1184,9 @@ DLLEXPORT char * freecell_solver_user_current_state_as_string(
 
         return
             fc_solve_state_as_string(
-#ifdef FCS_RCS_STATES
                 &(user->running_state.s),
                 &(user->running_state.info),
-#else
-                &(user->running_state),
-#endif
-#ifdef FCS_WITHOUT_LOCS_FIELDS
                 &(user->initial_state_locs),
-#endif
                 INSTANCE_FREECELLS_NUM,
                 INSTANCE_STACKS_NUM,
                 INSTANCE_DECKS_NUM,
@@ -1338,6 +1298,19 @@ void DLLEXPORT freecell_solver_user_set_solving_method(
 
     switch (method)
     {
+        case FCS_METHOD_RANDOM_DFS:
+        case FCS_METHOD_SOFT_DFS:
+        case FCS_METHOD_HARD_DFS:
+        {
+            user->soft_thread->method_specific.soft_dfs.dfs_max_depth
+                = user->soft_thread->method_specific.soft_dfs.depth
+                = user->soft_thread->method_specific.soft_dfs.tests_by_depth_array.num_units
+                = 0;
+            user->soft_thread->method_specific.soft_dfs.rand_seed = 24;
+            user->soft_thread->method_specific.soft_dfs.soft_dfs_info = NULL;
+            user->soft_thread->method_specific.soft_dfs.tests_by_depth_array.by_depth_units = NULL;
+        }
+        break;
         case FCS_METHOD_A_STAR:
         {
             double * my_befs_weights;
@@ -1527,7 +1500,7 @@ int DLLEXPORT freecell_solver_user_get_num_times(void * api_instance)
 
     user = (fcs_user_t *)api_instance;
 
-    return user->iterations_board_started_at + user->fc_solve_obj->num_times - user->init_num_times;
+    return user->iterations_board_started_at.num_times + user->fc_solve_obj->num_times - user->init_num_times.num_times;
 }
 
 int DLLEXPORT freecell_solver_user_get_limit_iterations(void * api_instance)
@@ -1733,7 +1706,7 @@ int DLLEXPORT freecell_solver_user_set_empty_stacks_filled_by(
 
 int DLLEXPORT freecell_solver_user_set_a_star_weight(
     void * api_instance,
-    int index,
+    int my_index,
     double weight
     )
 {
@@ -1742,7 +1715,7 @@ int DLLEXPORT freecell_solver_user_set_a_star_weight(
     user = (fcs_user_t *)api_instance;
 
 #define my_befs_weights soft_thread->method_specific.befs.meth.befs.befs_weights
-    if ((index < 0) || (index >= (sizeof(user->my_befs_weights)/sizeof(user->my_befs_weights[0]))))
+    if ((my_index < 0) || (my_index >= (sizeof(user->my_befs_weights)/sizeof(user->my_befs_weights[0]))))
     {
         return 1;
     }
@@ -1751,7 +1724,7 @@ int DLLEXPORT freecell_solver_user_set_a_star_weight(
         return 2;
     }
 
-    user->my_befs_weights[index] = weight;
+    user->my_befs_weights[my_index] = weight;
 
     return 0;
 
@@ -1761,7 +1734,7 @@ int DLLEXPORT freecell_solver_user_set_a_star_weight(
 #ifdef FCS_COMPILE_DEBUG_FUNCTIONS
 double DLLEXPORT fc_solve_user_INTERNAL_get_befs_weight(
     void * api_instance,
-    int index
+    int my_index
     )
 {
     fcs_user_t * user;
@@ -1770,7 +1743,7 @@ double DLLEXPORT fc_solve_user_INTERNAL_get_befs_weight(
 
 #define my_befs_weights soft_thread->method_specific.befs.meth.befs.befs_weights
 
-    return user->my_befs_weights[index];
+    return user->my_befs_weights[my_index];
 }
 
 #endif
@@ -1780,10 +1753,7 @@ static void iter_handler_wrapper(
     int iter_num,
     int depth,
     void * lp_instance GCC_UNUSED,
-#ifdef FCS_RCS_STATES
-    fcs_state_t * ptr_state_key,
-#endif
-    fcs_collectible_state_t * ptr_state_val,
+    fcs_kv_state_t * ptr_state,
     int parent_iter_num
     )
 {
@@ -1792,60 +1762,15 @@ static void iter_handler_wrapper(
 
     user = (fcs_user_t *)api_instance;
 
-    state_raw.s = ptr_state_val;
-#ifdef FCS_RCS_STATES
-    state_raw.key = ptr_state_key;
-    state_raw.val = ptr_state_val;
-#endif
-#ifdef FCS_WITHOUT_LOCS_FIELDS
-    {
-        int i;
-        for ( i=0 ; i<MAX_NUM_STACKS ; i++)
-        {
-            state_raw.locs.stack_locs[i] = (fcs_locs_t)i;
-        }
-        for ( i=0 ; i<MAX_NUM_FREECELLS ; i++)
-        {
-            state_raw.locs.fc_locs[i] = (fcs_locs_t)i;
-        }
-    }
-#endif
+    state_raw.key = ptr_state->key;
+    state_raw.val = ptr_state->val;
+    fc_solve_init_locs(&(state_raw.locs));
 
-#ifdef DEBUG
-    {
-        fc_solve_instance_t * instance = user->instance;
-        HT_LOOP_DECLARE_VARS();
-
-        HT_LOOP_START()
-        {
-            ST_LOOP_DECLARE_VARS();
-
-            ST_LOOP_START()
-            {
-                if (!strcmp(soft_thread->name, "11"))
-                {
-                    double * w =
-                        soft_thread->method_specific.befs
-                            .meth.befs.befs_weights
-                        ;
-
-                    printf("BeFS-Weights[\"11\"]=(%f,%f,%f,%f,%f)\n",
-                            w[0], w[1], w[2], w[3], w[4]
-                          );
-
-                    goto myend;
-                }
-            }
-        }
-
-    }
-myend:
-#endif
     user->iter_handler(
         api_instance,
         iter_num,
         depth,
-        (void *)&state_raw,
+        (void *)(&state_raw),
         parent_iter_num,
         user->iter_handler_context
         );
@@ -1853,19 +1778,10 @@ myend:
     return;
 }
 
-void set_debug_iter_output_func_to_val(
+static void set_debug_iter_output_func_to_val(
     fcs_user_t * user,
-        void (*value)(
-    void * api_instance,
-    int iter_num,
-    int depth,
-    void * lp_instance GCC_UNUSED,
-#ifdef FCS_RCS_STATES
-    fcs_state_t * ptr_state_key,
-#endif
-    fcs_collectible_state_t * ptr_state_val,
-    int parent_iter_num
-    ))
+    fcs_instance_debug_iter_output_func_t value
+)
 {
     FLARES_LOOP_DECLARE_VARS();
     FLARES_LOOP_START()
@@ -1923,15 +1839,9 @@ DLLEXPORT char * freecell_solver_user_iter_state_as_string(
 
     return
         fc_solve_state_as_string(
-#ifdef FCS_RCS_STATES
             ((fcs_standalone_state_ptrs_t *)ptr_state_void)->key,
             ((fcs_standalone_state_ptrs_t *)ptr_state_void)->val,
-#else
-            ((fcs_standalone_state_ptrs_t *)ptr_state_void)->s,
-#endif
-#ifdef FCS_WITHOUT_LOCS_FIELDS
             &(((fcs_standalone_state_ptrs_t *)ptr_state_void)->locs),
-#endif
             INSTANCE_FREECELLS_NUM,
             INSTANCE_STACKS_NUM,
             INSTANCE_DECKS_NUM,
@@ -1962,7 +1872,7 @@ int DLLEXPORT freecell_solver_user_get_num_states_in_collection(void * api_insta
 
     user = (fcs_user_t *)api_instance;
 
-    return user->fc_solve_obj->num_states_in_collection;
+    return user->iterations_board_started_at.num_states_in_collection + user->fc_solve_obj->num_states_in_collection - user->init_num_times.num_states_in_collection;
 }
 
 void DLLEXPORT freecell_solver_user_limit_num_states_in_collection(
@@ -2196,8 +2106,15 @@ void DLLEXPORT freecell_solver_user_recycle(
     {
         recycle_instance(user, i);
     }
+    /*
+     * Removing because we are still interested to keep the current iterations
+     * limit.
+     * */
+#if 0
     user->current_iterations_limit = -1;
-    user->iterations_board_started_at = 0;
+#endif
+    user->iterations_board_started_at.num_times = 0;
+    user->iterations_board_started_at.num_states_in_collection = 0;
     if (user->state_string_copy != NULL)
     {
         free(user->state_string_copy);
@@ -2240,7 +2157,7 @@ int DLLEXPORT freecell_solver_user_set_optimization_scan_tests_order(
 }
 
 
-int DLLEXPORT extern freecell_solver_user_set_pruning(
+extern int DLLEXPORT freecell_solver_user_set_pruning(
     void * api_instance,
     const char * pruning,
     char * * error_string
