@@ -26,7 +26,7 @@
  * executable.
  *
  * It is documented in the documents "README", "USAGE", etc. in the
- * Freecell Solver distribution from http://fc-solve.berlios.de/ .
+ * Freecell Solver distribution from http://fc-solve.shlomifish.org/ .
  */
 #include <string.h>
 #include <stdio.h>
@@ -42,47 +42,8 @@
 #include <unistd.h>
 #include <sys/types.h>
 #endif
-enum STANDARD_NOTATION_TYPE
-{
-    STANDARD_NOTATION_NO,
-    STANDARD_NOTATION_REGULAR,
-    STANDARD_NOTATION_EXTENDED
-};
 
-
-struct fc_solve_display_information_context_struct
-{
-    fcs_bool_t debug_iter_state_output;
-    int freecells_num;
-    int stacks_num;
-    int decks_num;
-    fcs_bool_t parseable_output;
-    fcs_bool_t canonized_order_output;
-    fcs_bool_t display_10_as_t;
-    fcs_bool_t display_parent_iter_num;
-    fcs_bool_t debug_iter_output_on;
-    fcs_bool_t display_moves;
-    fcs_bool_t display_states;
-    int standard_notation;
-    const char * output_filename;
-};
-
-typedef struct fc_solve_display_information_context_struct fc_solve_display_information_context_t;
-
-static void init_debug_context(
-    fc_solve_display_information_context_t * dc
-    )
-{
-    dc->debug_iter_state_output = FALSE;
-    dc->parseable_output = FALSE;
-    dc->canonized_order_output = FALSE;
-    dc->display_10_as_t = FALSE;
-    dc->display_parent_iter_num = FALSE;
-    dc->display_moves = FALSE;
-    dc->display_states = TRUE;
-    dc->standard_notation = STANDARD_NOTATION_NO;
-    dc->output_filename = NULL;
-}
+#include "output_to_file.h"
 
 static void my_iter_handler(
     void * user_instance,
@@ -359,7 +320,7 @@ help_screen_t help_screens[] = {
 "\n"
 "\n"
 "Freecell Solver was written by Shlomi Fish.\n"
-"Homepage: http://fc-solve.berlios.de/\n"
+"Homepage: http://fc-solve.shlomifish.org/\n"
 "Send comments and suggestions to http://www.shlomifish.org/me/contact-me/\n"
 },
 {
@@ -434,6 +395,8 @@ static void print_help_string(const char * key)
     }
 }
 
+#define IS_ARG(s) (!strcmp(arg_str, (s)))
+
 static int cmd_line_callback(
     void * instance,
     int argc,
@@ -444,12 +407,17 @@ static int cmd_line_callback(
     void * context
     )
 {
+    /* TODO: extract a macro IS_ARG("string") for IS_ARG("string")
+     * */
     fc_solve_display_information_context_t * dc;
+    const char * arg_str;
+
     *num_to_skip = 0;
 
     dc = (fc_solve_display_information_context_t * )context;
+    arg_str = argv[arg];
 
-    if (!strcmp(argv[arg], "--version"))
+    if (IS_ARG("--version"))
     {
         printf(
             "fc-solve\nlibfreecell-solver version %s\n",
@@ -458,9 +426,9 @@ static int cmd_line_callback(
         *ret = EXIT_AND_RETURN_0;
         return FCS_CMD_LINE_STOP;
     }
-    else if ((!strcmp(argv[arg], "-h")) || (!strcmp(argv[arg], "--help")))
+    else if (IS_ARG("-h") || IS_ARG("--help"))
     {
-        char * help_key;
+        const char * help_key;
 
         help_key = getenv("FREECELL_SOLVER_DEFAULT_HELP");
         if (help_key == NULL)
@@ -471,13 +439,13 @@ static int cmd_line_callback(
         *ret = EXIT_AND_RETURN_0;
         return FCS_CMD_LINE_STOP;
     }
-    else if (!strncmp(argv[arg], "--help-", 7))
+    else if (!strncmp(arg_str, "--help-", 7))
     {
-        print_help_string(argv[arg]+7);
+        print_help_string(arg_str+7);
         *ret = EXIT_AND_RETURN_0;
         return FCS_CMD_LINE_STOP;
     }
-    else if ((!strcmp(argv[arg], "-i")) || (!strcmp(argv[arg], "--iter-output")))
+    else if (IS_ARG("-i") || IS_ARG("--iter-output"))
     {
 #define set_iter_handler() \
         freecell_solver_user_set_iter_handler(   \
@@ -489,21 +457,21 @@ static int cmd_line_callback(
 
         set_iter_handler();
     }
-    else if ((!strcmp(argv[arg], "-s")) || (!strcmp(argv[arg], "--state-output")))
+    else if (IS_ARG("-s") || IS_ARG("--state-output"))
     {
         set_iter_handler();
         dc->debug_iter_state_output = TRUE;
 #undef set_iter_handler
     }
-    else if ((!strcmp(argv[arg], "-p")) || (!strcmp(argv[arg], "--parseable-output")))
+    else if (IS_ARG("-p") || IS_ARG("--parseable-output"))
     {
         dc->parseable_output = TRUE;
     }
-    else if ((!strcmp(argv[arg], "-c")) || (!strcmp(argv[arg], "--canonized-order-output")))
+    else if (IS_ARG("-c") || IS_ARG("--canonized-order-output"))
     {
         dc->canonized_order_output = TRUE;
     }
-    else if ((!strcmp(argv[arg], "-o")) || (!strcmp(argv[arg], "--output")))
+    else if (IS_ARG("-o") || IS_ARG("--output"))
     {
         arg++;
         if (arg == argc)
@@ -514,34 +482,38 @@ static int cmd_line_callback(
         dc->output_filename = (const char *)argv[arg];
         return FCS_CMD_LINE_SKIP;
     }
-    else if ((!strcmp(argv[arg], "-t")) || (!strcmp(argv[arg], "--display-10-as-t")))
+    else if (IS_ARG("-t") || IS_ARG("--display-10-as-t"))
     {
         dc->display_10_as_t = TRUE;
     }
-    else if ((!strcmp(argv[arg], "-m")) || (!strcmp(argv[arg], "--display-moves")))
+    else if (IS_ARG("-m") || IS_ARG("--display-moves"))
     {
         dc->display_moves = TRUE;
         dc->display_states = FALSE;
     }
-    else if ((!strcmp(argv[arg], "-sn")) || (!strcmp(argv[arg], "--standard-notation")))
+    else if (IS_ARG("-sn") || IS_ARG("--standard-notation"))
     {
         dc->standard_notation = STANDARD_NOTATION_REGULAR;
 
     }
-    else if ((!strcmp(argv[arg], "-snx")) || (!strcmp(argv[arg], "--standard-notation-extended")))
+    else if (IS_ARG("-snx") || IS_ARG("--standard-notation-extended"))
     {
         dc->standard_notation = STANDARD_NOTATION_EXTENDED;
     }
-    else if ((!strcmp(argv[arg], "-sam")) || (!strcmp(argv[arg], "--display-states-and-moves")))
+    else if (IS_ARG("-sam") || IS_ARG("--display-states-and-moves"))
     {
         dc->display_moves = TRUE;
         dc->display_states = TRUE;
     }
-    else if ((!strcmp(argv[arg], "-pi")) || (!strcmp(argv[arg], "--display-parent-iter")))
+    else if (IS_ARG("-pi") || IS_ARG("--display-parent-iter"))
     {
         dc->display_parent_iter_num = TRUE;
     }
-    else if ((!strcmp(argv[arg], "--reset")))
+    else if (IS_ARG("-sel") || IS_ARG("--show-exceeded-limits"))
+    {
+        dc->show_exceeded_limits = TRUE;
+    }
+    else if (IS_ARG("--reset"))
     {
         init_debug_context(dc);
         freecell_solver_user_set_iter_handler(
@@ -554,13 +526,14 @@ static int cmd_line_callback(
     }
     else
     {
-        printf("Unimplemented option - \"%s\"!", argv[arg]);
+        printf("Unimplemented option - \"%s\"!", arg_str);
         exit(-1);
     }
     *num_to_skip = 1;
     return FCS_CMD_LINE_SKIP;
 }
 
+#undef IS_ARG
 
 static int command_num = 0;
 static int debug_iter_output_on = FALSE;
@@ -628,6 +601,7 @@ static freecell_solver_str_t known_parameters[] = {
     "-snx", "--standard-notation-extended",
     "-sam", "--display-states-and-moves",
     "-pi", "--display-parent-iter",
+    "-sel", "--show-exceeded-limits",
     "-o", "--output",
     "--reset",
     "--version",
@@ -662,7 +636,7 @@ int main(int argc, char * argv[])
             freecell_solver_user_cmd_line_parse_args(
                 instance,
                 argc,
-                (freecell_solver_str_t *)argv,
+                (freecell_solver_str_t *)(void *)argv,
                 1,
                 (freecell_solver_str_t *)known_parameters,
                 cmd_line_callback,
@@ -676,17 +650,13 @@ int main(int argc, char * argv[])
             freecell_solver_user_free(instance);
             return 0;
         }
-        else if (
-            (parser_ret == FCS_CMD_LINE_PARAM_WITH_NO_ARG)
-                )
+        else if (parser_ret == FCS_CMD_LINE_PARAM_WITH_NO_ARG)
         {
             fprintf(stderr, "The command line parameter \"%s\" requires an argument"
                     " and was not supplied with one.\n", argv[arg]);
             return (-1);
         }
-        else if (
-            (parser_ret == FCS_CMD_LINE_ERROR_IN_ARG)
-            )
+        else if (parser_ret == FCS_CMD_LINE_ERROR_IN_ARG)
         {
             if (error_string != NULL)
             {
@@ -775,7 +745,7 @@ int main(int argc, char * argv[])
         while (ret == FCS_STATE_SUSPEND_PROCESS)
         {
             FCS_GET_TIME(mytime);
-            
+
             /* This was taken from:
              *
              * http://www.brokestream.com/procstat.html
@@ -876,125 +846,9 @@ int main(int argc, char * argv[])
             output_fh = stdout;
         }
 
-        if (ret == FCS_STATE_WAS_SOLVED)
-        {
-            fprintf(output_fh, "-=-=-=-=-=-=-=-=-=-=-=-\n\n");
-            {
-                fcs_move_t move;
-                FILE * move_dump;
-                char * as_string;
-                int move_num = 0;
-
-                move_dump = output_fh;
-
-                if (debug_context.display_states)
-                {
-                    as_string =
-                        freecell_solver_user_current_state_as_string(
-                            instance,
-                            debug_context.parseable_output,
-                            debug_context.canonized_order_output,
-                            debug_context.display_10_as_t
-                            );
-
-                    fprintf(move_dump, "%s\n", as_string);
-
-                    free(as_string);
-
-                    fprintf(move_dump, "%s", "\n====================\n\n");
-                }
-
-                while (
-                        freecell_solver_user_get_next_move(
-                            instance,
-                            &move
-                            ) == 0
-                        )
-                {
-                    if (debug_context.display_moves)
-                    {
-                        as_string =
-                            freecell_solver_user_move_to_string_w_state(
-                                instance,
-                                move,
-                                debug_context.standard_notation
-                                );
-
-                        if (debug_context.display_states && debug_context.standard_notation)
-                        {
-                            fprintf(move_dump, "Move: ");
-                        }
-
-                        fprintf(
-                            move_dump,
-                            (debug_context.standard_notation ?
-                                "%s " :
-                                "%s\n"
-                            ),
-                            as_string
-                            );
-                        move_num++;
-                        if (debug_context.standard_notation)
-                        {
-                            if ((move_num % 10 == 0) || debug_context.display_states)
-                            {
-                                fprintf(move_dump, "\n");
-                            }
-                        }
-                        if (debug_context.display_states)
-                        {
-                            fprintf(move_dump, "\n");
-                        }
-                        fflush(move_dump);
-                        free(as_string);
-                    }
-
-                    if (debug_context.display_states)
-                    {
-                        as_string =
-                            freecell_solver_user_current_state_as_string(
-                                instance,
-                                debug_context.parseable_output,
-                                debug_context.canonized_order_output,
-                                debug_context.display_10_as_t
-                                );
-
-                        fprintf(move_dump, "%s\n", as_string);
-
-                        free(as_string);
-                    }
-
-                    if (debug_context.display_states || (!debug_context.standard_notation))
-                    {
-                        fprintf(move_dump, "%s", "\n====================\n\n");
-                    }
-                }
-
-                if (debug_context.standard_notation && (!debug_context.display_states))
-                {
-                    fprintf(move_dump, "\n\n");
-                }
-            }
-
-            fprintf(output_fh, "This game is solveable.\n");
-        }
-        else
-        {
-            fprintf (output_fh, "I could not solve this game.\n");
-        }
-
-        fprintf(
-            output_fh,
-            "Total number of states checked is %i.\n",
-            freecell_solver_user_get_num_times(instance)
-            );
-#if 1
-        fprintf(
-            output_fh,
-            "This scan generated %i states.\n",
-            freecell_solver_user_get_num_states_in_collection(instance)
-            );
-#endif
+        fc_solve_output_result_to_file(
+            output_fh, instance, ret, &debug_context
+        );
 
         if (debug_context.output_filename)
         {
