@@ -44,6 +44,7 @@ extern "C" {
 #include "card.h"
 
 #include "internal_move_struct.h"
+#include "indirect_buffer.h"
 
 #if MAX_NUM_INITIAL_CARDS_IN_A_STACK+12>(MAX_NUM_DECKS*52)
 #define MAX_NUM_CARDS_IN_A_STACK (MAX_NUM_DECKS*52)
@@ -85,10 +86,6 @@ struct fcs_struct_state_t
     fc_stack_t stacks[MAX_NUM_STACKS];
     fcs_card_t freecells[MAX_NUM_FREECELLS];
     fcs_state_foundation_t foundations[MAX_NUM_DECKS*4];
-#ifdef FCS_WITH_TALONS
-    fcs_card_t * talon;
-    char talon_params[4];
-#endif
 };
 
 typedef struct fcs_struct_state_t fcs_state_t;
@@ -132,21 +129,6 @@ typedef int fcs_locs_t;
     (card).flags = (flipped)
 #endif
 
-#ifdef FCS_WITH_TALONS
-#define fcs_talon_len(state) \
-    ((state).talon_params[0])
-
-#define fcs_talon_pos(state) \
-    ((state).talon_params[1])
-
-#define fcs_get_talon_card(state, pos) \
-    ((state).talon[pos])
-
-#define fcs_put_card_in_talon(state, pos, card) \
-    ((state).talon[pos] = (card))
-#endif
-
-
 #elif defined(COMPACT_STATES)    /* #ifdef DEBUG_STATES */
 
 
@@ -163,11 +145,8 @@ typedef fcs_card_t fcs_state_foundation_t;
 struct fcs_struct_state_t
 {
     fcs_card_t data[MAX_NUM_STACKS*(MAX_NUM_CARDS_IN_A_STACK+1)+MAX_NUM_FREECELLS+4*MAX_NUM_DECKS];
-#ifdef FCS_WITH_TALONS
-    fcs_card_t * talon;
-    char talon_params[4];
-#endif
 };
+
 /*
  * Stack: 0 - Number of cards
  *        1-19 - Cards
@@ -206,20 +185,6 @@ typedef char fcs_locs_t;
 #define fcs_foundation_value(state, d) \
     ( (state).data[FCS_FOUNDATIONS_OFFSET+(d)])
 
-#ifdef FCS_WITH_TALONS
-#define fcs_talon_len(state) \
-    ((state).talon_params[0])
-
-#define fcs_talon_pos(state) \
-    ((state).talon_params[1])
-
-#define fcs_put_card_in_talon(state, pos, card) \
-    ((state).talon[pos] = (card))
-
-#define fcs_get_talon_card(state, pos) \
-    ((state).talon[pos])
-#endif
-
 #elif defined(INDIRECT_STACK_STATES) /* #ifdef DEBUG_STATES
                                         #elif defined(COMPACT_STATES)
                                       */
@@ -233,10 +198,6 @@ struct fcs_struct_state_t
     fcs_cards_column_t stacks[MAX_NUM_STACKS];
     fcs_card_t freecells[MAX_NUM_FREECELLS];
     fcs_state_foundation_t foundations[MAX_NUM_DECKS*4];
-#ifdef FCS_WITH_TALONS
-    fcs_card_t * talon;
-    char talon_params[4];
-#endif
 };
 
 typedef struct fcs_struct_state_t fcs_state_t;
@@ -252,30 +213,17 @@ typedef struct fcs_struct_state_t fcs_state_t;
 #define fcs_foundation_value(state, d) \
     ( (state).foundations[(d)] )
 
-#ifdef FCS_WITH_TALONS
-#define fcs_talon_len(state) \
-    ((state).talon_params[0])
 
-#define fcs_talon_pos(state) \
-    ((state).talon_params[1])
-
-#define fcs_put_card_in_talon(state, pos, card) \
-    ((state).talon[pos] = (card))
-
-#define fcs_get_talon_card(state, pos) \
-    ((state).talon[pos])
-#endif
-
-#define fcs_copy_stack(state_struct, idx, buffer) \
+#define fcs_copy_stack(state_key, state_val, idx, buffer) \
     {     \
-        if (! ((state_struct).info.stacks_copy_on_write_flags & (1 << idx)))        \
+        if (! ((state_val).stacks_copy_on_write_flags & (1 << idx)))        \
         {          \
             fcs_cards_column_t copy_stack_col; \
                                     \
-            (state_struct).info.stacks_copy_on_write_flags |= (1 << idx);       \
-            copy_stack_col = fcs_state_get_col(((state_struct).s), idx); \
+            (state_val).stacks_copy_on_write_flags |= (1 << idx);       \
+            copy_stack_col = fcs_state_get_col((state_key), idx); \
             memcpy(&buffer[idx << 7], copy_stack_col, fcs_col_len(copy_stack_col)+1); \
-            fcs_state_get_col((state_struct.s), idx) = &buffer[idx << 7];     \
+            fcs_state_get_col((state_key), idx) = &buffer[idx << 7];     \
         }     \
     }
 
@@ -283,6 +231,10 @@ typedef struct fcs_struct_state_t fcs_state_t;
     {   \
         (ptr_dest)->info.stacks_copy_on_write_flags = 0; \
     }
+#define fcs_duplicate_kv_state_extra(ptr_dest, ptr_src) \
+{ \
+    ptr_dest->stacks_copy_on_write_flags = 0; \
+}
 
 typedef char fcs_locs_t;
 
@@ -356,27 +308,15 @@ typedef char fcs_locs_t;
 #define fcs_col_push_col_card(dest_col, src_col, card_idx) \
     fcs_col_push_card((dest_col), fcs_col_get_card((src_col), (card_idx)))
 
-#ifdef FCS_RCS_STATES
-#define fcs_duplicate_state(ptr_dest_key, ptr_dest_val, ptr_src_key, ptr_src_val) \
-    { \
-    *(ptr_dest_key) = *(ptr_src_key); \
-    *(ptr_dest_val) = *(ptr_src_val); \
-    fcs_duplicate_state_extra(ptr_dest, ptr_src);   \
-    }
-#else
-#define fcs_duplicate_state(ptr_dest, ptr_src) \
-    { \
-    *(ptr_dest) = *(ptr_src); \
-    fcs_duplicate_state_extra(ptr_dest, ptr_src);   \
-    }
-#endif
-
 #if defined(COMPACT_STATES) || defined(DEBUG_STATES)
 
 #define fcs_duplicate_state_extra(ptr_dest, ptr_src) \
     {}
 
-#define fcs_copy_stack(state_struct, idx, buffer) {}
+#define fcs_duplicate_kv_state_extra(ptr_dest, ptr_src) \
+    {}
+
+#define fcs_copy_stack(state_key, state_val, idx, buffer) {}
 
 #endif
 
@@ -402,10 +342,10 @@ typedef char fcs_locs_t;
     ((col)[(card_idx)+1])
 
 #define fcs_card_card_num(card) \
-    ( (card) & 0x0F )
+    ( (card) >> 2 )
 
 #define fcs_card_suit(card) \
-    ( ((card) >> 4) & 0x03 )
+    ( (card) & 0x03 )
 
 #ifndef FCS_WITHOUT_CARD_FLIPPING
 #define fcs_card_get_flipped(card) \
@@ -413,10 +353,10 @@ typedef char fcs_locs_t;
 #endif
 
 #define fcs_card_set_num(card, num) \
-    (card) = ((fcs_card_t)(((card)&0xF0)|(num)));
+    (card) = ((fcs_card_t)(((card)&0x03)|((num)<<2)));
 
 #define fcs_card_set_suit(card, suit) \
-    (card) = ((fcs_card_t)(((card)&0x4F)|((suit)<<4)));
+    (card) = ((fcs_card_t)(((card)&0xFC)|(suit)));
 
 #ifndef FCS_WITHOUT_CARD_FLIPPING
 #define fcs_card_set_flipped(card, flipped) \
@@ -504,18 +444,8 @@ struct fcs_state_extra_info_struct
      * */
     int stacks_copy_on_write_flags;
 #endif
-#ifndef FCS_WITHOUT_LOCS_FIELDS
-    /*
-     * These contain the location of the original stacks and freecells
-     * in the permutation of them. They are sorted by the canonization
-     * function.
-     * */
-    fcs_locs_t stack_locs[MAX_NUM_STACKS];
-    fcs_locs_t fc_locs[MAX_NUM_FREECELLS];
-#endif
 };
 
-#ifdef FCS_WITHOUT_LOCS_FIELDS
 typedef struct
 {
     /*
@@ -526,7 +456,19 @@ typedef struct
     fcs_locs_t stack_locs[MAX_NUM_STACKS];
     fcs_locs_t fc_locs[MAX_NUM_FREECELLS];
 } fcs_state_locs_struct_t;
-#endif
+
+static GCC_INLINE void fc_solve_init_locs(fcs_state_locs_struct_t * locs)
+{
+    int i;
+    for ( i=0 ; i<MAX_NUM_STACKS ; i++)
+    {
+        locs->stack_locs[i] = (fcs_locs_t)i;
+    }
+    for ( i=0 ; i<MAX_NUM_FREECELLS ; i++)
+    {
+        locs->fc_locs[i] = (fcs_locs_t)i;
+    }
+}
 
 typedef struct fcs_state_extra_info_struct fcs_state_extra_info_t;
 
@@ -545,6 +487,11 @@ struct fcs_state_keyval_pair_struct
 
 typedef struct fcs_state_keyval_pair_struct fcs_state_keyval_pair_t;
 
+typedef struct {
+    fcs_state_t * key;
+    fcs_state_extra_info_t * val;
+} fcs_kv_state_t;
+
 /*
  * This type is the struct that is collectible inside the hash.
  *
@@ -554,25 +501,68 @@ typedef struct fcs_state_keyval_pair_struct fcs_state_keyval_pair_t;
  *
  * */
 #ifdef FCS_RCS_STATES
+
 typedef fcs_state_extra_info_t fcs_collectible_state_t;
 #define FCS_S_ACCESSOR(s, field) ((s)->field)
-#define FCS_S_NEXT(s) FCS_S_ACCESSOR(s, parent)
+
+#define fcs_duplicate_state(ptr_dest, ptr_src) \
+    { \
+    *((ptr_dest)->key) = *((ptr_src)->key); \
+    *((ptr_dest)->val) = *((ptr_src)->val); \
+    fcs_duplicate_state_extra(((ptr_dest)->val), ((ptr_src)->val));   \
+    }
+
+#define fcs_duplicate_kv_state(x,y) fcs_duplicate_state(x,y)
+
+#define FCS_STATE_keyval_pair_to_collectible(s) (&((s)->info))
+#define FCS_STATE_kv_to_collectible(s) ((s)->val)
+#define FCS_STATE_kv_to_pass(s) (s)
+
+static GCC_INLINE void
+FCS_STATE_collectible_to_kv(fcs_kv_state_t * ret, fcs_collectible_state_t * s)
+{
+    ret->val = s;
+}
+
 #else
+
 typedef fcs_state_keyval_pair_t fcs_collectible_state_t;
+
 #define FCS_S_ACCESSOR(s, field) (((s)->info).field)
-#define FCS_S_NEXT(s) ((s)->next)
+
+#define fcs_duplicate_state(ptr_dest, ptr_src) \
+    { \
+    *(ptr_dest) = *(ptr_src); \
+    fcs_duplicate_state_extra(ptr_dest, ptr_src);   \
+    }
+
+
+#define fcs_duplicate_kv_state(ptr_dest, ptr_src) \
+    { \
+    *((ptr_dest)->key) = *((ptr_src)->key); \
+    *((ptr_dest)->val) = *((ptr_src)->val); \
+    fcs_duplicate_kv_state_extra(((ptr_dest)->val), ((ptr_src)->val));   \
+    }
+
+#define FCS_STATE_keyval_pair_to_collectible(s) (s)
+#define FCS_STATE_kv_to_collectible(s) ((fcs_collectible_state_t *)((s)->key))
+#define FCS_STATE_kv_to_pass(s) FCS_STATE_kv_to_collectible(s)
+
+static GCC_INLINE void
+FCS_STATE_collectible_to_kv(fcs_kv_state_t * ret, fcs_collectible_state_t * s)
+{
+    ret->key = &(s->s);
+    ret->val = &(s->info);
+}
+
 #endif
 
+#define FCS_S_NEXT(s) FCS_S_ACCESSOR(s, parent)
 #define FCS_S_PARENT(s) FCS_S_ACCESSOR(s, parent)
 #define FCS_S_NUM_ACTIVE_CHILDREN(s) FCS_S_ACCESSOR(s, num_active_children)
 #define FCS_S_MOVES_TO_PARENT(s) FCS_S_ACCESSOR(s, moves_to_parent)
 #define FCS_S_VISITED(s) FCS_S_ACCESSOR(s, visited)
 
-#ifndef FCS_WITHOUT_LOCS_FIELDS
-#define FCS_S_STACK_LOCS(s) FCS_S_ACCESSOR(s, stack_locs)
-#define FCS_S_FC_LOCS(s) FCS_S_ACCESSOR(s, fc_locs)
-#endif
-
 #define FCS_S_SCAN_VISITED(s) FCS_S_ACCESSOR(s, scan_visited)
 
 #ifndef FCS_WITHOUT_DEPTH_FIELD
@@ -587,64 +577,22 @@ typedef struct {
     fcs_state_t * key;
     fcs_state_extra_info_t * val;
     fcs_collectible_state_t * s;
-#ifdef FCS_WITHOUT_LOCS_FIELDS
     fcs_state_locs_struct_t locs;
-#endif
 } fcs_standalone_state_ptrs_t;
 
 extern fcs_card_t fc_solve_empty_card;
 
-#ifdef FCS_WITH_TALONS
-#define fcs_klondike_talon_len(state) \
-    ((state).talon[0])
-
-#define fcs_klondike_talon_stack_pos(state) \
-    ((state).talon_params[0])
-
-#define fcs_klondike_talon_queue_pos(state) \
-    ((state).talon_params[1])
-
-#define fcs_klondike_talon_num_redeals_left(state) \
-    ((state).talon_params[2])
-
-#define fcs_klondike_talon_get_top_card(state) \
-    ((state).talon[(int)fcs_klondike_talon_stack_pos(state)])
-
-#define fcs_klondike_talon_queue_to_stack(state) \
-    ( ((state).talon[(int)((++fcs_klondike_talon_stack_pos(state))+1)]) =  \
-      ((state).talon[(int)((fcs_klondike_talon_queue_pos(state)++)+1)]) )
-
-#define fcs_klondike_talon_redeal_bare(state) \
-    {         \
-        fcs_klondike_talon_stack_pos(state) = -1; \
-        fcs_klondike_talon_queue_pos(state) = 0; \
-    }
-
-#define fcs_klondike_talon_decrement_stack(state) \
-    ((state).talon[(int)((fcs_klondike_talon_stack_pos(state)--)+1)] = fc_solve_empty_card)
-#endif
-
 extern void fc_solve_canonize_state(
-#ifdef FCS_RCS_STATES
-    fcs_state_t * state_key,
-#endif
-    fcs_collectible_state_t * state,
+    fcs_kv_state_t * state_raw,
     int freecells_num,
     int stacks_num
     );
 
-#ifdef FCS_WITHOUT_LOCS_FIELDS
 void fc_solve_canonize_state_with_locs(
-#ifdef FCS_RCS_STATES
-    fcs_state_t * state_key,
-    fcs_collectible_state_t * state_val,
-#else
-    fcs_collectible_state_t * state,
-#endif
+    fcs_kv_state_t * state,
     fcs_state_locs_struct_t * locs,
     int freecells_num,
     int stacks_num);
-#endif
 
 #if (FCS_STATE_STORAGE != FCS_STATE_STORAGE_INDIRECT)
 
@@ -670,11 +618,6 @@ extern int fc_solve_state_compare_with_context(const void * s1, const void * s2,
 extern int fc_solve_state_compare_indirect(const void * s1, const void * s2);
 extern int fc_solve_state_compare_indirect_with_context(const void * s1, const void * s2, void * context);
 #endif
-
-#ifdef FCS_WITH_TALONS
-extern int fcs_talon_compare_with_context(const void * s1, const void * s2, fcs_compare_context_t context);
-#endif
-
 enum
 {
     FCS_USER_STATE_TO_C__SUCCESS = 0,
@@ -733,30 +676,27 @@ extern int fc_solve_u2p_card_number(const char * string);
  * */
 extern int fc_solve_u2p_suit(const char * deck);
 
-static GCC_INLINE void fc_solve_state_init(
+#ifdef INDIRECT_STACK_STATES
+#define fc_solve_state_init(state, stacks_num, indirect_stacks_buffer) \
+    fc_solve_state_init_proto(state, stacks_num, indirect_stacks_buffer)
+#else
+#define fc_solve_state_init(state, stacks_num, indirect_stacks_buffer) \
+    fc_solve_state_init_proto(state, stacks_num)
+#endif
+
+static GCC_INLINE void fc_solve_state_init_proto(
     fcs_state_keyval_pair_t * state,
     int stacks_num
-#ifdef INDIRECT_STACK_STATES
-    ,char * indirect_stacks_buffer
-#endif
+    IND_BUF_T_PARAM(indirect_stacks_buffer)
     )
 {
-#if ((!defined(FCS_WITHOUT_CARD_FLIPPING)) || (!defined(FCS_WITHOUT_LOCS_FIELDS)) || defined(INDIRECT_STACK_STATES))
+#if ((!defined(FCS_WITHOUT_CARD_FLIPPING)) || defined(INDIRECT_STACK_STATES))
     int i;
 #endif
 
 
     memset(&(state->s), 0, sizeof(state->s));
 
-#ifndef FCS_WITHOUT_LOCS_FIELDS
-    {
-        fcs_locs_t * stack_locs = state->info.stack_locs;
-        for ( i=0 ; i<MAX_NUM_STACKS ; i++)
-        {
-            stack_locs[i] = (fcs_locs_t)i;
-        }
-    }
-#endif
 #ifdef INDIRECT_STACK_STATES
     for(i=0;i<stacks_num;i++)
     {
@@ -767,12 +707,6 @@ static GCC_INLINE void fc_solve_state_init(
     {
         state->s.stacks[i] = NULL;
     }
-#endif
-#ifndef FCS_WITHOUT_LOCS_FIELDS
-    for ( i=0 ; i<MAX_NUM_FREECELLS ; i++)
-    {
-        state->info.fc_locs[i] = (char)i;
-    }
 #endif
     state->info.parent = NULL;
     state->info.moves_to_parent = NULL;
@@ -790,31 +724,28 @@ static GCC_INLINE void fc_solve_state_init(
 #endif
 }
 
-static const char * const fc_solve_freecells_prefixes[] = { "FC:", "Freecells:", "Freecell:", ""};
-
-static const char * const fc_solve_foundations_prefixes[] = { "Decks:", "Deck:", "Founds:", "Foundations:", "Foundation:", "Found:", ""};
-
-#ifdef FCS_WITH_TALONS
-static const char * const fc_solve_talon_prefixes[] = { "Talon:", "Queue:" , ""};
-static const char * const fc_solve_num_redeals_prefixes[] = { "Num-Redeals:", "Readels-Num:", "Readeals-Number:", ""};
-#endif
+static const char * const fc_solve_freecells_prefixes[] = { "FC:", "Freecells:", "Freecell:", NULL};
 
+static const char * const fc_solve_foundations_prefixes[] = { "Decks:", "Deck:", "Founds:", "Foundations:", "Foundation:", "Found:", NULL};
 #ifdef WIN32
 #define strncasecmp(a,b,c) (strnicmp((a),(b),(c)))
 #endif
 
-static GCC_INLINE int fc_solve_initial_user_state_to_c(
+#ifdef INDIRECT_STACK_STATES
+#define fc_solve_initial_user_state_to_c(string, out_state, freecells_num, stacks_num, decks_num, indirect_stacks_buffer) \
+    fc_solve_initial_user_state_to_c_proto(string, out_state, freecells_num, stacks_num, decks_num, indirect_stacks_buffer)
+#else
+#define fc_solve_initial_user_state_to_c(string, out_state, freecells_num, stacks_num, decks_num, indirect_stacks_buffer) \
+    fc_solve_initial_user_state_to_c_proto(string, out_state, freecells_num, stacks_num, decks_num)
+#endif
+
+static GCC_INLINE int fc_solve_initial_user_state_to_c_proto(
     const char * string,
     fcs_state_keyval_pair_t * out_state,
     int freecells_num,
     int stacks_num,
     int decks_num
-#ifdef FCS_WITH_TALONS
-    ,int talon_type
-#endif
-#ifdef INDIRECT_STACK_STATES
-    , char * indirect_stacks_buffer
-#endif
+    IND_BUF_T_PARAM(indirect_stacks_buffer)
     )
 {
     int s,c;
@@ -823,17 +754,13 @@ static GCC_INLINE int fc_solve_initial_user_state_to_c(
     fcs_cards_column_t col;
     int first_line;
 
-    int prefix_found;
-    const char * const * prefixes;
-    int i;
+    const char * const * prefix;
     int decks_index[4];
 
     fc_solve_state_init(
         out_state,
-        stacks_num
-#ifdef INDIRECT_STACK_STATES
-        , indirect_stacks_buffer
-#endif
+        stacks_num,
+        indirect_stacks_buffer
         );
     str = string;
 
@@ -849,13 +776,6 @@ static GCC_INLINE int fc_solve_initial_user_state_to_c(
         } \
     }
 
-#ifdef FCS_WITH_TALONS
-    if (talon_type == FCS_TALON_KLONDIKE)
-    {
-        fcs_klondike_talon_num_redeals_left(ret) = -1;
-    }
-#endif
-
     for(s=0;s<stacks_num;s++)
     {
         /* Move to the next stack */
@@ -868,21 +788,19 @@ static GCC_INLINE int fc_solve_initial_user_state_to_c(
             }
             str++;
         }
+
         first_line = 0;
 
-        prefixes = fc_solve_freecells_prefixes;
-        prefix_found = 0;
-        for(i=0;prefixes[i][0] != '\0'; i++)
+        for (prefix = fc_solve_freecells_prefixes ; (*prefix) ; prefix++)
         {
-            if (!strncasecmp(str, prefixes[i], strlen(prefixes[i])))
+            if (!strncasecmp(str, (*prefix), strlen(*prefix)))
             {
-                prefix_found = 1;
-                str += strlen(prefixes[i]);
+                str += strlen(*prefix);
                 break;
             }
         }
 
-        if (prefix_found)
+        if (*prefix)
         {
             for(c=0;c<freecells_num;c++)
             {
@@ -937,19 +855,16 @@ static GCC_INLINE int fc_solve_initial_user_state_to_c(
             continue;
         }
 
-        prefixes = fc_solve_foundations_prefixes;
-        prefix_found = 0;
-        for(i=0;prefixes[i][0] != '\0'; i++)
+        for(prefix = fc_solve_foundations_prefixes ; (*prefix) ; prefix++)
         {
-            if (!strncasecmp(str, prefixes[i], strlen(prefixes[i])))
+            if (!strncasecmp(str, (*prefix), strlen(*prefix)))
             {
-                prefix_found = 1;
-                str += strlen(prefixes[i]);
+                str += strlen(*prefix);
                 break;
             }
         }
 
-        if (prefix_found)
+        if (*prefix)
         {
             int d;
 
@@ -972,7 +887,9 @@ static GCC_INLINE int fc_solve_initial_user_state_to_c(
                 str++;
                 while (*str == '-')
                     str++;
-                c = fc_solve_u2p_card_number(str);
+                /* Workaround for fc_solve_u2p_card_number's willingness
+                 * to designate the string '0' as 10. */
+                c = ((str[0] == '0') ? 0 : fc_solve_u2p_card_number(str));
                 while (
                         (*str != ' ') &&
                         (*str != '\t') &&
@@ -995,117 +912,6 @@ static GCC_INLINE int fc_solve_initial_user_state_to_c(
             continue;
         }
 
-#ifdef FCS_WITH_TALONS
-        prefixes = fc_solve_talon_prefixes;
-        prefix_found = 0;
-        for(i=0;prefixes[i][0] != '\0'; i++)
-        {
-            if (!strncasecmp(str, prefixes[i], strlen(prefixes[i])))
-            {
-                prefix_found = 1;
-                str += strlen(prefixes[i]);
-                break;
-            }
-        }
-
-        if (prefix_found)
-        {
-            /* Input the Talon */
-            int talon_size;
-
-            talon_size = MAX_NUM_DECKS*52+16;
-            ret.talon = malloc(sizeof(fcs_card_t)*talon_size);
-            fcs_talon_pos(ret) = 0;
-
-            for(c=0 ; c < talon_size ; c++)
-            {
-                /* Move to the next card */
-                if (c!=0)
-                {
-                    while(
-                        ((*str) != ' ') &&
-                        ((*str) != '\t') &&
-                        ((*str) != '\n') &&
-                        ((*str) != '\r')
-                    )
-                    {
-                        handle_eos();
-                        str++;
-                    }
-                    if ((*str == '\n') || (*str == '\r'))
-                    {
-                        break;
-                    }
-                }
-
-                while ((*str == ' ') || (*str == '\t'))
-                {
-                    str++;
-                }
-
-                if ((*str == '\n') || (*str == '\r'))
-                {
-                    break;
-                }
-
-                card = fc_solve_card_user2perl(str);
-
-                fcs_put_card_in_talon(ret, c+(talon_type==FCS_TALON_KLONDIKE), card);
-            }
-            fcs_talon_len(ret) = c;
-
-            if (talon_type == FCS_TALON_KLONDIKE)
-            {
-                int talon_len;
-
-                talon_len = fcs_talon_len(ret);
-                fcs_klondike_talon_len(ret) = talon_len;
-                fcs_klondike_talon_stack_pos(ret) = -1;
-                fcs_klondike_talon_queue_pos(ret) = 0;
-            }
-
-            s--;
-            continue;
-        }
-
-        prefixes = fc_solve_num_redeals_prefixes;
-        prefix_found = 0;
-        for(i=0;prefixes[i][0] != '\0'; i++)
-        {
-            if (!strncasecmp(str, prefixes[i], strlen(prefixes[i])))
-            {
-                prefix_found = 1;
-                str += strlen(prefixes[i]);
-                break;
-            }
-        }
-
-        if (prefix_found)
-        {
-            while ((*str < '0') && (*str > '9') && (*str != '\n'))
-            {
-                handle_eos();
-                str++;
-            }
-            if (*str != '\n')
-            {
-                int num_redeals;
-
-                num_redeals = atoi(str);
-                if (talon_type == FCS_TALON_KLONDIKE)
-                {
-                    fcs_klondike_talon_num_redeals_left(ret) =
-                        (num_redeals < 0) ?
-                            (-1) :
-                            ((num_redeals > 127) ? 127 : num_redeals)
-                                ;
-                }
-            }
-            s--;
-            continue;
-        }
-#endif
-
         col = fcs_state_get_col(ret, s);
         for(c=0 ; c < MAX_NUM_CARDS_IN_A_STACK ; c++)
         {
@@ -1150,13 +956,9 @@ static GCC_INLINE int fc_solve_initial_user_state_to_c(
 #undef handle_eos
 
 extern char * fc_solve_state_as_string(
-#ifdef FCS_RCS_STATES
     fcs_state_t * key,
-#endif
-    fcs_collectible_state_t * state_pair,
-#ifdef FCS_WITHOUT_LOCS_FIELDS
+    fcs_state_extra_info_t * state_val,
     fcs_state_locs_struct_t * state_locs,
-#endif
     int freecells_num,
     int stacks_num,
     int decks_num,
@@ -1179,9 +981,6 @@ static GCC_INLINE int fc_solve_check_state_validity(
     int freecells_num,
     int stacks_num,
     int decks_num,
-#ifdef FCS_WITH_TALONS
-    int talon_type,
-#endif
     fcs_card_t * misplaced_card)
 {
     int cards[4][14];
@@ -1243,24 +1042,6 @@ static GCC_INLINE int fc_solve_check_state_validity(
         }
     }
 
-#ifdef FCS_WITH_TALONS
-    /* Mark the cards in the (gypsy) talon */
-    if ((talon_type == FCS_TALON_GYPSY) || (talon_type == FCS_TALON_KLONDIKE))
-    {
-        for(c = ((talon_type == FCS_TALON_GYPSY)?fcs_talon_pos(*state):1) ;
-            c < ((talon_type==FCS_TALON_GYPSY) ? fcs_talon_len(*state) : (fcs_klondike_talon_len(*state)+1)) ;
-            c++)
-        {
-            if (fcs_get_talon_card(*state,c) != fc_solve_empty_card)
-            {
-                cards
-                    [fcs_card_suit(fcs_get_talon_card(*state, c))]
-                    [fcs_card_card_num(fcs_get_talon_card(*state, c))] ++;
-            }
-        }
-    }
-#endif
-
     /* Now check if there are extra or missing cards */
 
     for(d=0;d<4;d++)
@@ -1303,11 +1084,16 @@ enum
 #define min(a,b) ((a)<(b)?(a):(b))
 #endif
 
+#define FCS_WITH_CARD_COMPARE_LOOKUP
+
 static GCC_INLINE int fc_solve_card_compare(
-        const fcs_card_t * c1,
-        const fcs_card_t * c2
+        const fcs_card_t * const c1,
+        const fcs_card_t * const c2
         )
 {
+#ifdef FCS_WITH_CARD_COMPARE_LOOKUP
+    return (*c1)-(*c2);
+#else
     if (fcs_card_card_num(*c1) > fcs_card_card_num(*c2))
     {
         return 1;
@@ -1331,25 +1117,26 @@ static GCC_INLINE int fc_solve_card_compare(
             return 0;
         }
     }
+#endif
 }
 
 #if defined(INDIRECT_STACK_STATES)
-static GCC_INLINE int fc_solve_stack_compare_for_comparison(const void * v_s1, const void * v_s2)
+static GCC_INLINE int fc_solve_stack_compare_for_comparison(const void * const v_s1, const void * const v_s2)
 {
-    const fcs_card_t * s1 = (const fcs_card_t *)v_s1;
-    const fcs_card_t * s2 = (const fcs_card_t *)v_s2;
+    const fcs_card_t * const s1 = (const fcs_card_t * const)v_s1;
+    const fcs_card_t * const s2 = (const fcs_card_t * const)v_s2;
 
-    int min_len;
-    int a, ret;
-
-    min_len = min(s1[0], s2[0]);
-
-    for(a=0;a<min_len;a++)
+    const int min_len = min(s1[0], s2[0]);
     {
-        ret = fc_solve_card_compare(s1+a+1,s2+a+1);
-        if (ret != 0)
+        int a, ret;
+
+        for(a=0;a<min_len;a++)
         {
-            return ret;
+            ret = fc_solve_card_compare(s1+a+1,s2+a+1);
+            if (ret != 0)
+            {
+                return ret;
+            }
         }
     }
     /*
@@ -1358,6 +1145,7 @@ static GCC_INLINE int fc_solve_stack_compare_for_comparison(const void * v_s1, c
      * fcs_stack_compare_for_stack_sort, and with the one card comparison
      * of the other state representation mechanisms.
      * */
+    /* For some reason this code is faster than s1[0]-s2[0] */
     if (s1[0] < s2[0])
     {
         return -1;
